"""The py21cmfast package."""

from importlib.metadata import PackageNotFoundError, version

try:
    from ._version import version as __version__
except ModuleNotFoundError:  # pragma: no cover
    try:
        __version__ = version("21cmFAST")
    except PackageNotFoundError:
        # package is not installed
        __version__ = "unknown"

__all__ = [
    "DATA_PATH",
    "AngularLightconer",
    "AstroOptions",
    "AstroParams",
    "BrightnessTemp",
    "CacheConfig",
    "Coeval",
    "CosmoParams",
    "GlobalEvolution",
    "HaloBox",
    "HaloCatalog",
    "InitialConditions",
    "InputParameters",
    "IonizedBox",
    "LightCone",
    "MatterOptions",
    "OutputCache",
    "PerturbedField",
    "PerturbedHaloCatalog",
    "RectilinearLightconer",
    "RunCache",
    "SimulationOptions",
    "TsBox",
    "XraySourceBox",
    "__version__",
    "brightness_temperature",
    "compute_halo_grid",
    "compute_initial_conditions",
    "compute_ionization_field",
    "compute_luminosity_function",
    "compute_rms",
    "compute_spin_temperature",
    "compute_tau",
    "compute_xray_source_field",
    "config",
    "configure_logging",
    "construct_fftw_wisdoms",
    "create_params_from_template",
    "determine_halo_catalog",
    "generate_coeval",
    "generate_lightcone",
    "get_logspaced_redshifts",
    "lightconers",
    "list_templates",
    "perturb_field",
    "perturb_halo_catalog",
    "plotting",
    "read_inputs",
    "read_output_struct",
    "run_classy",
    "run_coeval",
    "run_global_evolution",
    "run_lightcone",
    "setup_photon_cons",
    "wrapper",
    "write_output_to_hdf5",
    "write_template",
]

from . import lightconers, plotting, wrapper
from ._cfg import config
from ._data import DATA_PATH
from ._logging import configure_logging
from ._templates import create_params_from_template, list_templates, write_template
from .drivers.coeval import Coeval, generate_coeval, run_coeval
from .drivers.global_evolution import GlobalEvolution, run_global_evolution
from .drivers.lightcone import LightCone, generate_lightcone, run_lightcone
from .drivers.single_field import (
    brightness_temperature,
    compute_halo_grid,
    compute_initial_conditions,
    compute_ionization_field,
    compute_spin_temperature,
    compute_xray_source_field,
    determine_halo_catalog,
    perturb_field,
    perturb_halo_catalog,
)
from .io.caching import CacheConfig, OutputCache, RunCache
from .io.h5 import read_inputs, read_output_struct, write_output_to_hdf5
from .lightconers import AngularLightconer, RectilinearLightconer
from .wrapper.cfuncs import (
    compute_luminosity_function,
    compute_tau,
    construct_fftw_wisdoms,
)
from .wrapper.classy_interface import compute_rms, run_classy
from .wrapper.inputs import (
    AstroOptions,
    AstroParams,
    CosmoParams,
    InputParameters,
    MatterOptions,
    SimulationOptions,
    get_logspaced_redshifts,
)
from .wrapper.outputs import (
    BrightnessTemp,
    HaloBox,
    HaloCatalog,
    InitialConditions,
    IonizedBox,
    PerturbedField,
    PerturbedHaloCatalog,
    TsBox,
    XraySourceBox,
)
from .wrapper.photoncons import setup_photon_cons

configure_logging()



"""Data structure wrappers for the C code."""

from __future__ import annotations

import contextlib
import logging
from typing import Any

import attrs
from bidict import bidict

from ..c_21cmfast import ffi
from .arrays import Array

logger = logging.getLogger(__name__)


@attrs.define(slots=False)
class StructWrapper:
    """
    A base-class python wrapper for C structures (not instances of them).

    Provides simple methods for creating new instances and accessing field names and values.

    To implement wrappers of specific structures, make a subclass with the same name as the
    appropriate C struct (which must be defined in the C code that has been compiled to the ``ffi``
    object), *or* use an arbitrary name, but set the ``_name`` attribute to the C struct name.
    """

    _name: str = attrs.field(converter=str)
    cstruct = attrs.field(default=None)
    _ffi = attrs.field(default=ffi)

    _TYPEMAP = bidict({"float32": "float *", "float64": "double *", "int32": "int *"})

    @_name.default
    def _name_default(self):
        return self.__class__.__name__

    def __init__(self, *args):
        """Perform custom initializion actions.

        This instantiates the memory associated with the C struct, attached to this inst.
        """
        self.__attrs_init__(*args)
        self.cstruct = self._new()

    def _new(self):
        """Return a new empty C structure corresponding to this class."""
        return self._ffi.new(f"struct {self._name}*")

    @property
    def fields(self) -> dict[str, Any]:
        """A list of fields of the underlying C struct (a list of tuples of "name, type")."""
        return dict(self._ffi.typeof(self.cstruct[0]).fields)

    @property
    def fieldnames(self) -> list[str]:
        """A list of names of fields of the underlying C struct."""
        return [f for f, t in self.fields.items()]

    @property
    def pointer_fields(self) -> list[str]:
        """A list of names of fields which have pointer type in the C struct."""
        return [f for f, t in self.fields.items() if t.type.kind == "pointer"]

    @property
    def primitive_fields(self) -> list[str]:
        """The list of names of fields which have primitive type in the C struct."""
        return [f for f, t in self.fields.items() if t.type.kind == "primitive"]

    def __getstate__(self):
        """Return the current state of the class without pointers."""
        return {
            k: v
            for k, v in self.__dict__.items()
            if k not in ["_strings", "cstruct", "_ffi"]
        }

    def expose_to_c(self, array: Array, name: str):
        """Expose the memory of a particular Array to the backend C code."""
        if not array.state.initialized:
            raise ValueError("Array must be initialized before exposing to C")

        def _ary2buf(ary):
            return self._ffi.cast(
                self._TYPEMAP[ary.dtype.name], self._ffi.from_buffer(ary)
            )

        try:
            setattr(self.cstruct, name, _ary2buf(array.value))
        except TypeError as e:
            raise TypeError(f"Error setting {name}") from e


class StructInstanceWrapper:
    """A wrapper for *instances* of C structs.

    This is as opposed to :class:`StructWrapper`, which is for the un-instantiated structs.

    Parameters
    ----------
    wrapped :
        The reference to the C object to wrap (contained in the ``cffi.lib`` object).
    ffi :
        The ``cffi.ffi`` object.
    """

    def __init__(self, wrapped, ffi):
        self._cobj = wrapped
        self._ffi = ffi

        for nm, _tp in self._ffi.typeof(self._cobj).fields:
            setattr(self, nm, getattr(self._cobj, nm))

        # Get the name of the structure
        self._ctype = self._ffi.typeof(self._cobj).cname.split()[-1]

    def __setattr__(self, name, value):
        """Set an attribute of the instance, attempting to change it in the C struct as well."""
        with contextlib.suppress(AttributeError):
            setattr(self._cobj, name, value)
        object.__setattr__(self, name, value)

    def items(self):
        """Yield (name, value) pairs for each element of the struct."""
        for nm, _tp in self._ffi.typeof(self._cobj).fields:
            yield nm, getattr(self, nm)

    def keys(self):
        """Return a list of names of elements in the struct."""
        return [nm for nm, tp in self.items()]

    def __iter__(self):
        """Iterate over the object like a dict."""
        yield from self.keys()

    def __repr__(self):
        """Return a unique representation of the instance."""
        return (
            self._ctype + "(" + ";".join(f"{k}={v!s}" for k, v in sorted(self.items()))
        ) + ")"

    def filtered_repr(self, filter_params):
        """Get a fully unique representation of the instance that filters out some parameters.

        Parameters
        ----------
        filter_params : list of str
            The parameter names which should not appear in the representation.
        """
        return (
            self._ctype
            + "("
            + ";".join(
                f"{k}={v!s}" for k, v in sorted(self.items()) if k not in filter_params
            )
        ) + ")"



"""Tests of miscellaneous utility functions not part of the main 21cmFAST simulation modules."""

import numpy as np
import pytest

from py21cmfast import InputParameters
from py21cmfast.utils import recursive_difference, show_references


def test_ref_printing():
    inputs = InputParameters.from_template("latest", random_seed=1234)
    ref_str = show_references(inputs, print_to_stdout=False)

    assert "2011MNRAS.411..955M" in ref_str
    assert "10.21105/joss.02582" in ref_str
    assert "10.1093/mnras/stz032" in ref_str
    assert "10.1093/mnras/staa1131" not in ref_str
    assert "10.1093/mnras/stac185" not in ref_str
    assert "10.48550/arXiv.2504.17254" not in ref_str

    inputs = InputParameters.from_template("mini-dhalos", random_seed=1234)
    ref_str = show_references(inputs, print_to_stdout=False)

    assert "2011MNRAS.411..955M" in ref_str
    assert "10.21105/joss.02582" in ref_str
    assert "10.1093/mnras/stz032" in ref_str
    assert "10.1093/mnras/staa1131" in ref_str
    assert "10.1093/mnras/stac185" in ref_str
    assert "10.48550/arXiv.2504.17254" in ref_str

    inputs = InputParameters.from_template("const-zeta", random_seed=1234)
    ref_str = show_references(inputs, print_to_stdout=True)

    assert "2011MNRAS.411..955M" in ref_str
    assert "10.21105/joss.02582" in ref_str
    assert "10.1093/mnras/stz032" not in ref_str
    assert "10.1093/mnras/staa1131" not in ref_str
    assert "10.1093/mnras/stac185" not in ref_str
    assert "10.48550/arXiv.2504.17254" not in ref_str


class TestRecursiveDifference:
    """Tests of the recursive_difference function."""

    def test_b_empty(self):
        """Test if b is empty in a - b gives a."""
        a = {"a": 1, "b": 2}
        aa = recursive_difference(a, {})

        assert a == aa

    def test_a_empty(self):
        """Test that {} - b == {}."""
        a = {}
        b = {"a": 1, "b": 2}
        aa = recursive_difference(a, b)
        assert len(aa) == 0

    def test_disjoint(self):
        """Test that a - b == a if b is disjoint with a."""
        a = {"a": 1}
        b = {"b": 2}
        assert recursive_difference(a, b) == a

    def test_a_recursive_b_not_recursive(self):
        """Test where a is recursive but b is not."""
        a = {"a": {"a": 1, "b": 2}, "b": 1}
        b = {"a": 1, "b": 2}

        assert recursive_difference(a, b) == a

    def test_a_and_b_recurse(self):
        """Test where both a and b are recursive."""
        a = {"a": {"a": 1, "b": 2}, "b": 1}
        b = a

        assert recursive_difference(a, b) == {}

    def test_comparison_rules(self):
        """Test that passing cmprules works as expected."""
        a = {"a": np.zeros(10)}
        b = {"a": np.zeros(10)}

        with pytest.raises(ValueError, match="The truth value of an array"):
            recursive_difference(a, b)

        cmprules = {np.ndarray: lambda x, y: np.allclose(x, y)}
        assert recursive_difference(a, b, cmprules=cmprules) == {}



"""Test high-level I/O."""

from pathlib import Path

import attrs
import h5py
import numpy as np
import pytest

from py21cmfast import (
    Coeval,
    CosmoParams,
    InitialConditions,
    InputParameters,
    LightCone,
    OutputCache,
    run_coeval,
    run_lightcone,
)
from py21cmfast.drivers.lightcone import AngularLightcone
from py21cmfast.io import h5
from py21cmfast.lightconers import AngularLightconer


@pytest.fixture(scope="module")
def coeval(ic, default_input_struct_ts, cache) -> Coeval:
    return run_coeval(
        out_redshifts=25.0,
        initial_conditions=ic,
        write=True,
        inputs=default_input_struct_ts,
        cache=cache,
    )[0]


@pytest.fixture(scope="module")
def ang_lightcone(ic, lc, default_input_struct_lc, cache):
    lcn = AngularLightconer.like_rectilinear(
        match_at_z=lc.lightcone_redshifts.min(),
        max_redshift=lc.lightcone_redshifts.max(),
        simulation_options=ic.simulation_options,
    )

    return run_lightcone(
        lightconer=lcn,
        initial_conditions=ic,
        write=True,
        inputs=default_input_struct_lc,
        include_dvdr_in_tau21=False,
        cache=cache,
    )


def test_read_bad_file_lc(test_direc: Path, lc: LightCone):
    # create a bad hdf5 file with some good fields,
    #  some bad fields, and some missing fields
    #  in both input parameters and box structures
    fname = test_direc / "_lc.h5"
    lc.save(path=fname)

    with h5py.File(fname, "r+") as f:
        # make gluts, these should be ignored on reading
        f["InputParameters"]["simulation_options"].attrs["NotARealParameter"] = (
            "fake_param"
        )

        # make gaps
        del f["InputParameters"]["cosmo_params"].attrs["hlittle"]

    # load without compatibility mode, make sure we throw the right error
    with pytest.raises(ValueError, match="Excess arguments exist"):
        LightCone.from_file(fname, safe=True)

    # load in compatibility mode, check that we warn correctly
    with pytest.warns(UserWarning, match="Excess arguments exist"):
        lc2 = LightCone.from_file(fname, safe=False)

    # check that the fake fields didn't show up in the struct
    assert not hasattr(lc2.simulation_options, "NotARealParameter")

    # check that missing fields are set to default
    assert lc2.cosmo_params.SIGMA_8 == CosmoParams().SIGMA_8

    # check that the fields which are good are read in the struct
    assert lc2.simulation_options == lc.simulation_options
    assert all(
        getattr(lc2.cosmo_params, field.name) == getattr(lc.cosmo_params, field.name)
        for field in attrs.fields(CosmoParams)
        if field.name != "SIGMA_8"
    )


def test_read_bad_file_coev(test_direc: Path, coeval: Coeval):
    # create a bad hdf5 file with some good fields,
    #  some bad fields, and some missing fields
    #  in both input parameters and box structures
    fname = test_direc / "_a_bad_file.h5"

    coeval.save(path=fname)
    with h5py.File(fname, "r+") as f:
        # make gluts, these should be ignored on reading
        f["BrightnessTemp"]["InputParameters"]["simulation_options"].attrs[
            "NotARealParameter"
        ] = "fake_param"

        # make gaps
        del f["BrightnessTemp"]["InputParameters"]["cosmo_params"].attrs["hlittle"]

    # load in the coeval check that we warn correctly
    with pytest.raises(ValueError, match="Excess arguments exist"):
        Coeval.from_file(fname, safe=True)

    with pytest.warns(UserWarning, match="Excess arguments exist"):
        cv2 = Coeval.from_file(fname, safe=False)

    # check that the fake params didn't show up in the struct
    assert not hasattr(cv2.simulation_options, "NotARealParameter")

    # check that missing fields are set to default
    assert cv2.cosmo_params.SIGMA_8 == CosmoParams().SIGMA_8

    # check that the fields which are good are read in the struct
    assert cv2.simulation_options == coeval.simulation_options
    assert all(
        getattr(cv2.cosmo_params, field.name)
        == getattr(coeval.cosmo_params, field.name)
        for field in attrs.fields(CosmoParams)
        if field.name != "SIGMA_8"
    )


def test_lightcone_roundtrip(test_direc, lc):
    fname = test_direc / "lc.h5"
    lc.save(path=fname)
    lc2 = LightCone.from_file(fname)

    assert lc == lc2
    assert np.allclose(lc.lightcone_redshifts, lc2.lightcone_redshifts)
    assert np.all(
        np.isclose(lc.lightcones["brightness_temp"], lc2.lightcones["brightness_temp"])
    )


def test_coeval_roundtrip(test_direc, coeval):
    fname = test_direc / "a_coeval.h5"
    coeval.save(fname)
    coeval2 = Coeval.from_file(fname)

    assert coeval == coeval2
    assert np.all(np.isclose(coeval.brightness_temp, coeval2.brightness_temp))


def test_ang_lightcone(lc, ang_lightcone: AngularLightcone, plt):
    # we test that the fields are "highly correlated",
    # and moreso in the one corner where the lightcones
    # should be almost exactly the same, and less so in the other
    # corners, and also less so at the highest redshifts.
    rbt = lc.lightcones["brightness_temp"]
    abt = ang_lightcone.lightcones["brightness_temp"].reshape(rbt.shape)

    fullcorr0 = np.corrcoef(rbt[:, :, 0].flatten(), abt[:, :, 0].flatten())
    fullcorrz = np.corrcoef(rbt[:, :, -1].flatten(), abt[:, :, -1].flatten())

    _fig, ax = plt.subplots(2, 2, figsize=(12, 12))
    ax[0, 0].imshow(rbt[:, :, 0])
    ax[0, 0].set_title("lowest rect")
    ax[0, 1].imshow(rbt[:, :, -1])
    ax[0, 1].set_title("highest rect")
    ax[1, 0].imshow(abt[:, :, 0])
    ax[1, 0].set_title("lowest ang")
    ax[1, 1].imshow(abt[:, :, -1])
    ax[1, 1].set_title("highest ang")

    print("correlation at low z: ", fullcorr0)
    print("correlation at highz: ", fullcorrz)
    assert fullcorr0[0, 1] > fullcorrz[0, 1]  # 0,0 and 1,1 are autocorrs.
    assert fullcorr0[0, 1] > 0.5

    # check corners
    n = rbt.shape[0]
    topcorner = np.corrcoef(
        rbt[: n // 2, : n // 2, 0].flatten(), abt[: n // 2, : n // 2, 0].flatten()
    )
    bottomcorner = np.corrcoef(
        rbt[n // 2 :, n // 2 :, 0].flatten(), abt[n // 2 :, n // 2 :, 0].flatten()
    )
    assert topcorner[0, 1] > bottomcorner[0, 1]


def test_write_to_group(ic: InitialConditions, cache: OutputCache):
    h5.write_output_to_hdf5(ic, path=cache.direc / "a_new_file.h5", group="new_group")

    with h5py.File(cache.direc / "a_new_file.h5", "r") as fl:
        assert "new_group" in fl

    ic2 = h5.read_output_struct(cache.direc / "a_new_file.h5", group="new_group")

    assert ic2 == ic


def test_write_output_to_hdf5_uncomputed(tmp_path):
    """Test that writing an uncomputed box to file raises an error."""
    inputs = InputParameters(random_seed=0)
    ics = InitialConditions.new(inputs)
    with pytest.raises(ValueError, match="Not all boxes have been computed"):
        h5.write_output_to_hdf5(ics, path=tmp_path / "ics.h5")


def read_nonexistent_output_struct(ic: InitialConditions, tmp_path: Path):
    """Test that attempting to read a non-existent struct from a file fails."""
    pth = tmp_path / "ics.h5"
    h5.write_output_to_hdf5(ic, pth)

    with pytest.raises(
        KeyError, match="struct non-existent not found in the HDF5 group"
    ):
        h5.read_output_struct(pth, struct="non-existent")


def test_read_inputs_from_filepath(tmp_path: Path):
    pth = tmp_path / "tmp.h5"
    inputs = InputParameters(random_seed=0)

    with h5py.File(pth, "w") as fl:
        h5._write_inputs_to_group(inputs, fl)

    new = h5.read_inputs(pth)  # directly read from path.
    assert new == inputs


def test_read_inputs_badfile(ic: InitialConditions, perturbed_field, tmp_path: Path):
    pth = tmp_path / "ics.h5"
    h5.write_output_to_hdf5(ic, pth)
    h5.write_output_to_hdf5(perturbed_field, pth, mode="a")

    with pytest.raises(h5.HDF5FileStructureError, match="Multiple sub-groups found"):
        h5.read_inputs(pth)


def test_read_inputs_version_warning(tmp_path: Path):
    pth = tmp_path / "tmp.h5"
    inputs = InputParameters(random_seed=0)

    with h5py.File(pth, "w") as fl:
        h5._write_inputs_to_group(inputs, fl)
        fl["InputParameters"].attrs["21cmFAST-version"] = "5.0.0"  # larger than reality

    with pytest.warns(UserWarning, match="File created with a newer version"):
        h5.read_inputs(pth)


def test_read_outputs_version_warning(ic: InitialConditions, tmp_path: Path):
    pth = tmp_path / "ics.h5"
    h5.write_output_to_hdf5(ic, pth)

    # Mock the version as newer than it really is.
    with h5py.File(pth, "a") as fl:
        fl["InitialConditions"]["OutputFields"].attrs["21cmFAST-version"] = "5.0.0"

    with pytest.warns(UserWarning, match="File created with a newer version"):
        h5.read_output_struct(pth)


def test_read_output_struct_missing_arrays(ic: InitialConditions, tmp_path: Path):
    pth = tmp_path / "ics.h5"
    h5.write_output_to_hdf5(ic, pth)

    # Artificially remove an array
    with h5py.File(pth, "a") as fl:
        del fl["InitialConditions"]["OutputFields"]["lowres_density"]

    with pytest.raises(
        h5.HDF5FileStructureError, match="Required Array lowres_density not found"
    ):
        h5.read_output_struct(pth)


def test_read_output_struct_badshape_arrays(ic: InitialConditions, tmp_path: Path):
    pth = tmp_path / "ics.h5"
    h5.write_output_to_hdf5(ic, pth)

    # Artificially remove an array
    with h5py.File(pth, "a") as fl:
        dens = fl["InitialConditions"]["OutputFields"]["lowres_density"][:10, :11, :12]
        del fl["InitialConditions"]["OutputFields"]["lowres_density"]
        fl["InitialConditions"]["OutputFields"]["lowres_density"] = dens

    with pytest.raises(
        h5.HDF5FileStructureError, match="Array lowres_density has shape"
    ):
        h5.read_output_struct(pth)


def test_read_inputs_non_existent_version(tmp_path: Path):
    pth = tmp_path / "tmp.h5"
    inputs = InputParameters(random_seed=0)

    with h5py.File(pth, "w") as fl:
        h5._write_inputs_to_group(inputs, fl)
        del fl["InputParameters"].attrs["21cmFAST-version"]

    with pytest.raises(NotImplementedError, match=f"The file {pth} is not a valid"):
        h5.read_inputs(pth)


def test_read_output_struct_no_version(ic: InitialConditions, tmp_path: Path):
    pth = tmp_path / "ics.h5"
    h5.write_output_to_hdf5(ic, pth)

    # Artificially remove an array
    with h5py.File(pth, "a") as fl:
        del fl["InitialConditions"]["OutputFields"].attrs["21cmFAST-version"]

    with pytest.raises(NotImplementedError, match=f"The file {pth} is not a valid"):
        h5.read_output_struct(pth)



"""Compute simulations that evolve over redshift."""

import logging
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import Self, get_args

import attrs
import h5py
import numpy as np
from rich import progress as prg
from rich.console import Console
from rich.progress import Progress

from .. import __version__
from ..c_21cmfast import lib
from ..io import h5
from ..io.caching import CacheConfig, OutputCache, RunCache
from ..rsds import apply_rsds, include_dvdr_in_tau21
from ..wrapper.arrays import Array
from ..wrapper.inputs import InputParameters
from ..wrapper.outputs import (
    BrightnessTemp,
    HaloBox,
    HaloCatalog,
    InitialConditions,
    IonizedBox,
    OutputStruct,
    PerturbedField,
    TsBox,
)
from ..wrapper.photoncons import _get_photon_nonconservation_data, setup_photon_cons
from . import single_field as sf
from ._param_config import high_level_func

logger = logging.getLogger(__name__)

_console = Console()


def _progressbar(**kwargs):
    return Progress(
        prg.TextColumn("[progress.description]{task.description:>24}"),
        prg.BarColumn(),
        prg.TaskProgressColumn(),
        prg.TextColumn("•"),
        prg.MofNCompleteColumn(),
        "[green]redshifts",
        prg.TextColumn("•"),
        prg.TimeElapsedColumn(),
        prg.TextColumn("•"),
        prg.TimeRemainingColumn(),
        "remaining",
        console=_console,
        **kwargs,
    )


@attrs.define
class Coeval:
    """A full coeval box with all associated data."""

    initial_conditions: InitialConditions = attrs.field(
        validator=attrs.validators.instance_of(InitialConditions)
    )
    perturbed_field: PerturbedField = attrs.field(
        validator=attrs.validators.instance_of(PerturbedField)
    )
    ionized_box: IonizedBox = attrs.field(
        validator=attrs.validators.instance_of(IonizedBox)
    )
    brightness_temperature: BrightnessTemp = attrs.field(
        validator=attrs.validators.instance_of(BrightnessTemp)
    )
    ts_box: TsBox = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(TsBox)),
    )
    halobox: HaloBox = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(HaloBox)),
    )
    photon_nonconservation_data: dict = attrs.field(factory=dict)

    def __getattr__(self, name):
        """
        Get underlying Array objects as attributes.

        This method allows accessing arrays from OutputStruct objects within the Coeval instance
        as if they were direct attributes of the Coeval object.

        Parameters
        ----------
        name : str
            The name of the attribute being accessed.

        Returns
        -------
        Any
            The value of the requested array from the appropriate OutputStruct object.

        Raises
        ------
        AttributeError
            If the requested attribute is not found in any of the OutputStruct objects.
        """
        # We only want to expose fields that are part of the Coeval object
        for box in attrs.asdict(self, recurse=False).values():
            if isinstance(box, OutputStruct) and name in box.arrays:
                return box.get(name)
        raise AttributeError(f"Coeval has no attribute '{name}'")

    @property
    def output_structs(self) -> dict[str, OutputStruct]:
        """
        Get a dictionary of OutputStruct objects contained in this Coeval instance.

        This property method returns a dictionary containing all the OutputStruct
        objects that are attributes of the Coeval instance. It filters out any
        non-OutputStruct attributes.

        Returns
        -------
        dict[str, OutputStruct]
            A dictionary where the keys are attribute names and the values are
            the corresponding OutputStruct objects.
        """
        return {
            k: v
            for k, v in attrs.asdict(self, recurse=False).items()
            if isinstance(v, OutputStruct)
        }

    @classmethod
    def get_fields(cls, ignore_structs: tuple[str] = ()) -> list[str]:
        """Obtain a list of name of simulation boxes saved in the Coeval object."""
        output_structs = []
        for fld in attrs.fields(cls):
            if fld.name in ignore_structs:
                continue

            if issubclass(fld.type, OutputStruct):
                output_structs.append(fld.type)
            else:
                args = get_args(fld.type)
                for k in args:
                    if issubclass(k, OutputStruct):
                        output_structs.append(k)
                        break

        pointer_fields = []
        for struct in output_structs:
            pointer_fields += [
                k for k, v in attrs.fields_dict(struct).items() if v.type == Array
            ]

        return pointer_fields

    @property
    def redshift(self) -> float:
        """The redshift of the coeval box."""
        return self.perturbed_field.redshift

    @property
    def inputs(self) -> InputParameters:
        """An InputParameters object associated with the coeval box."""
        return self.brightness_temperature.inputs

    @property
    def simulation_options(self):
        """Matter Params shared by all datasets."""
        return self.inputs.simulation_options

    @property
    def matter_options(self):
        """Matter Flags shared by all datasets."""
        return self.inputs.matter_options

    @property
    def cosmo_params(self):
        """Cosmo params shared by all datasets."""
        return self.inputs.cosmo_params

    @property
    def astro_options(self):
        """Flag Options shared by all datasets."""
        return self.inputs.astro_options

    @property
    def astro_params(self):
        """Astro params shared by all datasets."""
        return self.inputs.astro_params

    @property
    def random_seed(self):
        """Random seed shared by all datasets."""
        return self.inputs.random_seed

    def save(self, path: str | Path, clobber=False):
        """Save the Coeval object to disk."""
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        file_mode = "w" if clobber else "a"
        with h5py.File(path, file_mode) as fl:
            fl.attrs["coeval"] = True  # marker identifying this as a coeval box
            fl.attrs["__version__"] = __version__

            grp = fl.create_group("photon_nonconservation_data")
            for k, v in self.photon_nonconservation_data.items():
                grp[k] = v

        output_structs = self.output_structs
        for struct in output_structs.values():
            h5.write_output_to_hdf5(struct, path, mode="a")

    def include_dvdr_in_tau21(self, axis: str = "z", periodic: bool = True):
        """Include velocity gradient corrections to the brightness temperature field.

        Parameters
        ----------
        axis: str, optioanl
            The assumed axis of the line-of-sight. Options are "x", "y", or "z". Default is "z".
        periodic: bool, optioanl
            Whether to assume periodic boundary conditions along the line-of-sight. Default is True.

        Returns
        -------
        tb_with_dvdr
            A box of the brightness temperature, with velocity gradient corrections.
        """
        if not hasattr(self, "velocity_" + axis):
            if axis not in ["x", "y", "z"]:
                raise ValueError("`axis` can only be `x`, `y` or `z`.")
            else:
                raise ValueError(
                    "You asked for axis = '"
                    + axis
                    + "', but the coeval doesn't have velocity_"
                    + axis
                    + "!"
                    " Set matter_options.KEEP_3D_VELOCITIES=True next time you call run_coeval if you "
                    "wish to set axis=`" + axis + "'."
                )

        return include_dvdr_in_tau21(
            brightness_temp=self.brightness_temp,
            los_velocity=getattr(self, "velocity_" + axis),
            redshifts=self.redshift,  # TODO: do we want to use a single redshift? Or a redshift array that is determined from the coeval los?
            inputs=self.inputs,
            tau_21=self.tau_21 if self.inputs.astro_options.USE_TS_FLUCT else None,
            periodic=periodic,
        )

    def apply_rsds(
        self,
        field: str = "brightness_temp",
        axis: str = "z",
        periodic: bool = True,
        n_rsd_subcells: int = 4,
    ):
        """Apply redshift-space distortions to a particular field of the coeval box.

        Parameters
        ----------
        field: str, optional
            The field onto which redshift-space distortions shall be applied. Default is "brightness_temp".
        axis: str, optioanl
            The assumed axis of the line-of-sight. Options are "x", "y", or "z". Default is "z".
        periodic: bool, optioanl
            Whether to assume periodic boundary conditions along the line-of-sight. Default is True.
        n_rsd_subcells: int, optional
            The number of subcells into which each cell is divided when redshift space distortions are applied. Default is 4.

        Returns
        -------
        field_with_rsds
            A box of the field, with redshift space distortions.
        """
        if not hasattr(self, "velocity_" + axis):
            if axis not in ["x", "y", "z"]:
                raise ValueError("`axis` can only be `x`, `y` or `z`.")
            else:
                raise ValueError(
                    "You asked for axis = '"
                    + axis
                    + "', but the coeval doesn't have velocity_"
                    + axis
                    + "!"
                    " Set matter_options.KEEP_3D_VELOCITIES=True next time you call run_coeval if you "
                    "wish to set axis=`" + axis + "'."
                )

        return apply_rsds(
            field=getattr(self, field),
            los_velocity=getattr(self, "velocity_" + axis),
            redshifts=self.redshift,  # TODO: do we want to use a single redshift? Or a redshift array that is determined from the coeval los?
            inputs=self.inputs,
            periodic=periodic,
            n_rsd_subcells=n_rsd_subcells,
        )

    def apply_velocity_corrections(
        self, axis: str = "z", periodic: bool = True, n_rsd_subcells: int = 4
    ):
        """Apply velocity gradient corrections and redshift-space distortions to the brightness temperature field.

        Parameters
        ----------
        axis: str, optioanl
            The assumed axis of the line-of-sight. Options are "x", "y", or "z". Default is "z".
        periodic: bool, optioanl
            Whether to assume periodic boundary conditions along the line-of-sight. Default is True.
        n_rsd_subcells: int, optional
            The number of subcells into which each cell is divided when redshift space distortions are applied. Default is 4.

        Returns
        -------
        field_with_rsds
            A box of the brightness temperature, with velocity gradient corrections and redshift-space distortions.
        """
        if not hasattr(self, "velocity_" + axis):
            if axis not in ["x", "y", "z"]:
                raise ValueError("`axis` can only be `x`, `y` or `z`.")
            else:
                raise ValueError(
                    "You asked for axis = '"
                    + axis
                    + "', but the coeval doesn't have velocity_"
                    + axis
                    + "!"
                    " Set matter_options.KEEP_3D_VELOCITIES=True next time you call run_coeval if you "
                    "wish to set axis=`" + axis + "'."
                )

        tb_with_dvdr = include_dvdr_in_tau21(
            brightness_temp=self.brightness_temp,
            los_velocity=getattr(self, "velocity_" + axis),
            redshifts=self.redshift,  # TODO: do we want to use a single redshift? Or a redshift array that is determined from the coeval los?
            inputs=self.inputs,
            tau_21=self.tau_21 if self.inputs.astro_options.USE_TS_FLUCT else None,
            periodic=periodic,
        )

        return apply_rsds(
            field=tb_with_dvdr,
            los_velocity=getattr(self, "velocity_" + axis),
            redshifts=self.redshift,  # TODO: do we want to use a single redshift? Or a redshift array that is determined from the coeval los?
            inputs=self.inputs,
            periodic=periodic,
            n_rsd_subcells=n_rsd_subcells,
        )

    @classmethod
    def from_file(cls, path: str | Path, safe: bool = True) -> Self:
        """Read the Coeval object from disk and return it."""
        path = Path(path)
        if not path.exists():
            raise FileExistsError(f"The file {path} does not exist!")

        selfdict = attrs.fields_dict(cls)
        type_to_name = {v.type.__name__: k for k, v in selfdict.items()}

        with h5py.File(path, "r") as fl:
            if not fl.attrs.get("coeval", False):
                raise ValueError(f"The file {path} is not a Coeval file!")

            keys = set(fl.keys())

            grp = fl["photon_nonconservation_data"]
            photoncons = {k: v[...] for k, v in grp.items()}
            keys.remove("photon_nonconservation_data")

            kwargs = {
                type_to_name[k]: h5.read_output_struct(path, struct=k, safe=safe)
                for k in keys
            }
            return cls(photon_nonconservation_data=photoncons, **kwargs)

    def __eq__(self, other):
        """Determine if this is equal to another object."""
        return (
            isinstance(other, self.__class__)
            and other.inputs == self.inputs
            and self.redshift == other.redshift
        )


def evolve_halos(
    inputs: InputParameters,
    all_redshifts: list[float],
    write: CacheConfig,
    initial_conditions: InitialConditions,
    cache: OutputCache,
    regenerate: bool,
    free_cosmo_tables: bool,
    progressbar: bool = False,
):
    """
    Evolve and perturb halo fields across multiple redshifts.

    This function computes and evolves halo fields for a given set of redshifts,
    applying perturbations to each halo list. It processes redshifts in reverse order
    to account for descendant halos.

    Parameters
    ----------
    inputs : InputParameters
        Input parameters for the simulation.
    all_redshifts : list[float]
        List of redshifts to process, in descending order.
    write : CacheConfig
        Configuration for writing output to cache.
    initial_conditions : InitialConditions
        Initial conditions for the simulation.
    cache : OutputCache
        Cache object for storing and retrieving computed results.
    regenerate : bool
        Flag to indicate whether to regenerate results or use cached values.
    progressbar: bool, optional
        If True, a progress bar will be displayed throughout the simulation. Defaults to False.

    Returns
    -------
    list
        A list of perturbed halo fields for each redshift, in ascending redshift order.
        Returns an empty list if halo fields are not used or fixed grids are enabled.
    """
    # get the halos (reverse redshift order)
    if not inputs.matter_options.has_discrete_halos:
        return []

    if not write.halo_catalog and len(all_redshifts) > 1:
        warnings.warn(
            "You have turned off caching for the perturbed halo fields, but are"
            " evolving them across multiple redshifts. This will result in very high memory usage",
            stacklevel=2,
        )

    halofield_list = []
    kw = {
        "initial_conditions": initial_conditions,
        "cache": cache,
        "regenerate": regenerate,
        "free_cosmo_tables": free_cosmo_tables,
    }
    halos_desc = None
    with _progressbar(disable=not progressbar) as _progbar:
        for _i, z in _progbar.track(
            enumerate(all_redshifts[::-1]),
            description="Evolving Halos",
            total=len(all_redshifts),
        ):
            halos = sf.determine_halo_catalog(
                redshift=z,
                inputs=inputs,
                descendant_halos=halos_desc,
                write=write.halo_catalog,
                **kw,
            )

            halofield_list.append(halos)

            if z in inputs.node_redshifts:
                # Only evolve on the node_redshifts, not any redshifts in-between
                # that the user might care about.
                # we never want to store every halofield
                if write.halo_catalog and (halos_desc is not None):
                    halos_desc.purge()
                halos_desc = halos

    # reverse to get the right redshift order
    return halofield_list[::-1]


@high_level_func
def generate_coeval(
    *,
    inputs: InputParameters | None = None,
    out_redshifts: float | tuple[float] = (),
    regenerate: bool | None = None,
    write: CacheConfig | bool = True,
    cache: OutputCache | None = None,
    initial_conditions: InitialConditions | None = None,
    cleanup: bool = True,
    progressbar: bool = False,
):
    r"""
    Perform a full coeval simulation of all fields at given redshifts.

    This is generally the easiest and most efficient way to generate a set of coeval cubes at a
    given set of redshifts. It self-consistently deals with situations in which the field needs to be
    evolved, and does this with the highest memory-efficiency, only returning the desired redshift.
    All other calculations are by default stored in the on-disk cache so they can be re-used at a
    later time.

    Some calculations of the coeval quantities require redshift evolution, i.e. the
    calculation of higher-redshift coeval boxes up to some maximum redshift in order
    to integrate the quantities over cosmic time. The redshifts that define this
    evolution are set by the ``inputs.node_redshifts`` parameter. However, in some
    simple cases, this evolution is not required, and this parameter can be empty.
    Thus there is a distinction between the redshifts required for computing the physics
    (i.e. ``inputs.node_redshifts``) and the redshifts at which the user wants to
    obtain the resulting coeval cubes. The latter is controlled by ``out_redshifts``.
    If not set, ``out_redshifts`` will be set to ``inputs.node_redshifts``, so that
    all computed redshifts are returned as coeval boxes.

    .. note:: User-supplied ``out_redshifts`` are *not* used in the redshift evolution,
              so that the results depend precisely on the ``node_redshifts`` defined
              in the input parameters.

    Parameters
    ----------
    inputs: :class:`~InputParameters`
        This object specifies the input parameters for the run, including the random seed
    out_redshifts: array_like, optional
        A single redshift, or multiple redshifts, at which to return results. By default,
        use all the ``inputs.node_redshifts``. If neither is specified, an error will be
        raised.
    regenerate : bool
        If True, regenerate all fields, even if they are in the cache.
    write : :class:`~py21cmfast.cache.CacheConfig`, optional
        Either a bool specifying whether to write _all_ the boxes to cache (or none of
        them), or a :class:`~py21cmfast.cache.CacheConfig` object specifying which boxes
        to write.
    cache : :class:`~py21cmfast.cache.OutputCache`, optional
        The cache object to use for reading and writing data from the cache. This should
        be an instance of :class:`~py21cmfast.cache.OutputCache`, which depends solely
        on specifying a directory to host the cache.
    initial_conditions : :class:`~InitialConditions`, optional
        If given, use these intial conditions as a basis for computing the other
        fields, instead of re-computing the ICs. If this is defined, the ``inputs`` do
        not need to be defined (but can be, in order to overwrite the ``node_redshifts``).
    cleanup : bool, optional
        A flag to specify whether the C routine cleans up its memory before returning.
        Typically, if `spin_temperature` is called directly, you will want this to be
        true, as if the next box to be calculated has different shape, errors will occur
        if memory is not cleaned. Note that internally, this is set to False until the
        last iteration.
    progressbar: bool, optional
        If True, a progress bar will be displayed throughout the simulation. Defaults to False.

    Returns
    -------
    coevals : list of :class:`~py21cmfast.drivers.coeval.Coeval`
        The full data for the Coeval class, with init boxes, perturbed fields, ionized boxes,
        brightness temperature, and potential data from the conservation of photons. A
        list of such objects, one for each redshift in ``out_redshifts``.
    """
    if cache is None:
        cache = OutputCache(".")

    if isinstance(write, bool):
        write = CacheConfig() if write else CacheConfig.off()

    if not out_redshifts:
        out_redshifts = inputs.node_redshifts

    if not out_redshifts and not inputs.node_redshifts:
        raise ValueError("out_redshifts must be given if inputs has no node redshifts")

    iokw = {"regenerate": regenerate, "cache": cache, "free_cosmo_tables": False}

    if not hasattr(out_redshifts, "__len__"):
        out_redshifts = [out_redshifts]

    if isinstance(out_redshifts, np.ndarray):
        out_redshifts = out_redshifts.tolist()

    # Get the list of redshifts we need to scroll through.
    all_redshifts = _get_required_redshifts_coeval(inputs, out_redshifts)

    (
        initial_conditions,
        perturbed_field,
        halofield_list,
        photon_nonconservation_data,
    ) = _setup_ics_and_pfs_for_scrolling(
        all_redshifts=all_redshifts,
        inputs=inputs,
        initial_conditions=initial_conditions,
        write=write,
        progressbar=progressbar,
        **iokw,
    )

    # get the first node redshift that is above all output redshifts
    first_out_node = None
    if inputs.node_redshifts:
        # if any output redshift is above all nodes, start from the start
        if max(out_redshifts) > max(inputs.node_redshifts):
            first_out_node = -1
        # otherwise, find the first node above all user redshifts and start there
        elif max(out_redshifts) >= min(inputs.node_redshifts):
            nodes_in_outputs = np.array(inputs.node_redshifts) <= max(out_redshifts)
            first_out_node = np.where(nodes_in_outputs)[0][0] - 1
    idx, coeval = _obtain_starting_point_for_scrolling(
        inputs=inputs,
        initial_conditions=initial_conditions,
        photon_nonconservation_data=photon_nonconservation_data,
        cache=cache,
        regenerate=regenerate,
        minimum_node=first_out_node,
    )
    # convert node_redshift index to all_redshift index
    if idx > 0:
        idx = np.argmin(np.fabs(np.array(all_redshifts) - inputs.node_redshifts[idx]))

    for _, coeval in _redshift_loop_generator(  # noqa: B020
        inputs=inputs,
        all_redshifts=all_redshifts,
        initial_conditions=initial_conditions,
        photon_nonconservation_data=photon_nonconservation_data,
        perturbed_field=perturbed_field,
        halofield_list=halofield_list,
        write=write,
        cleanup=cleanup,
        progressbar=progressbar,
        iokw=iokw,
        init_coeval=coeval,
        start_idx=idx + 1,
    ):
        yield coeval, coeval.redshift in out_redshifts

    if lib.photon_cons_allocated:
        lib.FreePhotonConsMemory()

    lib.Free_cosmo_tables_global()


def run_coeval(**kwargs) -> list[Coeval]:
    """Run a coeval simulation and return the resulting coeval boxes.

    This simply wraps :func:`generate_coeval` and returns a list of coeval objects
    at the requested output redshifts after the generator has been exhausted. All
    parameters are passed directly to :func:`generate_coeval`.
    """
    return [coeval for coeval, in_outputs in generate_coeval(**kwargs) if in_outputs]


def _obtain_starting_point_for_scrolling(
    inputs: InputParameters,
    initial_conditions: InitialConditions,
    photon_nonconservation_data: dict,
    cache: OutputCache,
    regenerate: bool,
    minimum_node: int | None = None,
):
    outputs = None

    if minimum_node is None:
        # By default, check for completeness at all nodes, starting at
        # the last one.
        minimum_node = len(inputs.node_redshifts) - 1

    if minimum_node < 0 or inputs.matter_options.has_discrete_halos or regenerate:
        # TODO: (low priority) implement a backward loop for finding first halo files
        #   Noting that we need *all* the halo fields in the cache to run
        return (
            -1,
            None,
        )

    logger.info(f"Determining pre-cached boxes for the run in {cache}")
    rc = RunCache.from_inputs(inputs, cache)

    for idx in range(minimum_node, -1, -1):
        if not rc.is_complete_at(index=idx):
            continue

        _z = inputs.node_redshifts[idx]
        outputs = rc.get_all_boxes_at_z(z=_z)
        break

    # Create a Coeval from the outputs
    if outputs is not None:
        return idx, Coeval(
            initial_conditions=initial_conditions,
            perturbed_field=outputs["PerturbedField"],
            ionized_box=outputs["IonizedBox"],
            brightness_temperature=outputs["BrightnessTemp"],
            ts_box=outputs.get("TsBox", None),
            halobox=outputs.get("Halobox", None),
            photon_nonconservation_data=photon_nonconservation_data,
        )
    else:
        return -1, None


def _redshift_loop_generator(
    inputs: InputParameters,
    initial_conditions: InitialConditions,
    all_redshifts: Sequence[float],
    perturbed_field: list[PerturbedField],
    halofield_list: list[HaloCatalog],
    write: CacheConfig,
    iokw: dict,
    cleanup: bool,
    progressbar: bool,
    photon_nonconservation_data: dict,
    start_idx: int = 0,
    init_coeval: Coeval | None = None,
):
    if isinstance(write, bool):
        write = CacheConfig()

    # Iterate through redshift from top to bottom
    hbox_arr = []

    prev_coeval = init_coeval
    this_coeval = None

    this_halobox = None
    this_spin_temp = None
    this_halofield = None
    this_xraysource = None

    kw = {
        **iokw,
        "initial_conditions": initial_conditions,
    }
    with _progressbar(disable=not progressbar) as _progbar:
        for iz, z in _progbar.track(
            enumerate(all_redshifts),
            description="Evolving Astrophysics",
            total=len(all_redshifts),
        ):
            if not progressbar:
                logger.info(
                    f"Computing Redshift {z} ({iz + 1}/{len(all_redshifts)}) iterations."
                )
            if iz < start_idx:
                continue

            this_perturbed_field = perturbed_field[iz]
            this_perturbed_field.load_all()

            if inputs.matter_options.lagrangian_source_grid:
                if inputs.matter_options.has_discrete_halos:
                    this_halofield = halofield_list[iz]
                    this_halofield.load_all()
                this_halobox = sf.compute_halo_grid(
                    inputs=inputs,
                    halo_catalog=this_halofield,
                    redshift=z,
                    previous_ionize_box=getattr(prev_coeval, "ionized_box", None),
                    previous_spin_temp=getattr(prev_coeval, "ts_box", None),
                    write=write.halobox,
                    **kw,
                )

            if inputs.astro_options.USE_TS_FLUCT:
                if inputs.matter_options.lagrangian_source_grid:
                    # append the halo redshift array so we have all halo boxes [z,zmax]
                    this_xraysource = sf.compute_xray_source_field(
                        redshift=z,
                        hboxes=[*hbox_arr, this_halobox],
                        write=write.xray_source_box,
                        **kw,
                    )

                this_spin_temp = sf.compute_spin_temperature(
                    inputs=inputs,
                    previous_spin_temp=getattr(prev_coeval, "ts_box", None),
                    perturbed_field=this_perturbed_field,
                    xray_source_box=this_xraysource,
                    write=write.spin_temp,
                    **kw,
                    cleanup=(cleanup and z == all_redshifts[-1]),
                )
                # Purge XraySourceBox because it's enormous
                if inputs.matter_options.lagrangian_source_grid:
                    this_xraysource.purge(force=True)

            this_ionized_box = sf.compute_ionization_field(
                inputs=inputs,
                previous_ionized_box=getattr(prev_coeval, "ionized_box", None),
                perturbed_field=this_perturbed_field,
                # perturb field *not* interpolated here.
                previous_perturbed_field=getattr(prev_coeval, "perturbed_field", None),
                halobox=this_halobox,
                spin_temp=this_spin_temp,
                write=write.ionized_box,
                **kw,
            )

            this_bt = sf.brightness_temperature(
                ionized_box=this_ionized_box,
                perturbed_field=this_perturbed_field,
                spin_temp=this_spin_temp,
                write=write.brightness_temp,
                **iokw,
            )

            if inputs.astro_options.PHOTON_CONS_TYPE == "z-photoncons":
                # Updated info at each z.
                photon_nonconservation_data = _get_photon_nonconservation_data()

            this_coeval = Coeval(
                initial_conditions=initial_conditions,
                perturbed_field=this_perturbed_field,
                ionized_box=this_ionized_box,
                brightness_temperature=this_bt,
                ts_box=this_spin_temp,
                halobox=this_halobox,
                photon_nonconservation_data=photon_nonconservation_data,
            )

            # We purge previous fields and those we no longer need
            if prev_coeval is not None:
                if (
                    inputs.simulation_options.HII_DIM > 1
                ):  # No need to purge if we have just one cell
                    prev_coeval.perturbed_field.purge()
                if (
                    inputs.matter_options.lagrangian_source_grid
                    and write.halobox
                    and iz + 1 < len(all_redshifts)
                ):
                    for hbox in hbox_arr:
                        hbox.prepare_for_next_snapshot(next_z=all_redshifts[iz + 1])

            if this_halofield is not None:
                this_halofield.purge()

            if z in inputs.node_redshifts:
                # Only evolve on the node_redshifts, not any redshifts in-between
                # that the user might care about.
                prev_coeval = this_coeval
                hbox_arr += [this_halobox]

            # yield before the cleanup, so we can get at the fields before they are purged
            yield iz, this_coeval


def _setup_ics_and_pfs_for_scrolling(
    all_redshifts: Sequence[float],
    initial_conditions: InitialConditions | None,
    inputs: InputParameters,
    write: CacheConfig,
    progressbar: bool,
    overdensity_z0: float | None = None,
    **iokw,
) -> tuple[InitialConditions, list[PerturbedField], list[HaloCatalog], dict]:
    if initial_conditions is None:
        initial_conditions = sf.compute_initial_conditions(
            inputs=inputs,
            write=write.initial_conditions,
            initial_density=overdensity_z0,
            **iokw,
        )

    # We can go ahead and purge some of the stuff in the initial_conditions, but only if
    # it is cached -- otherwise we could be losing information.
    if write.initial_conditions:
        initial_conditions.prepare_for_perturb()
    kw = {
        "initial_conditions": initial_conditions,
        **iokw,
    }
    photon_nonconservation_data = {}
    if inputs.astro_options.PHOTON_CONS_TYPE != "no-photoncons":
        # Note that we need to pass the inputs directly here.
        # Otherwise, they are taken from initial_conditions, which may have
        # a different (compatible) set of inputs.
        photon_nonconservation_data = setup_photon_cons(inputs=inputs, **kw)

    if (
        inputs.astro_options.PHOTON_CONS_TYPE == "z-photoncons"
        and np.amin(all_redshifts) < inputs.astro_params.PHOTONCONS_CALIBRATION_END
    ):
        raise ValueError(
            f"You have passed a redshift (z = {np.amin(all_redshifts)}) that is lower than"
            "the endpoint of the photon non-conservation correction"
            f"(astro_params.PHOTONCONS_CALIBRATION_END = {inputs.astro_params.PHOTONCONS_CALIBRATION_END})."
            "If this behaviour is desired then set astro_params.PHOTONCONS_CALIBRATION_END"
            f"to a value lower than z = {np.amin(all_redshifts)}."
        )

    # Get all the perturb boxes early. We need to get the perturb at every
    # redshift.
    perturbed_field = []

    with _progressbar(disable=not progressbar) as _progbar:
        for z in _progbar.track(all_redshifts, description="Perturbing Matter Fields"):
            p = sf.perturb_field(
                redshift=z,
                inputs=inputs,
                write=write.perturbed_field,
                **kw,
            )

            if inputs.matter_options.MINIMIZE_MEMORY and write.perturbed_field:
                p.purge()
            perturbed_field.append(p)

    halofield_list = evolve_halos(
        inputs=inputs,
        all_redshifts=all_redshifts,
        write=write,
        progressbar=progressbar,
        **kw,
    )
    # Now we can purge initial_conditions further.
    if write.initial_conditions:
        initial_conditions.prepare_for_spin_temp()

    return (
        initial_conditions,
        perturbed_field,
        halofield_list,
        photon_nonconservation_data,
    )


def _get_required_redshifts_coeval(
    inputs: InputParameters, user_redshifts: Sequence
) -> list[float]:
    # Add in the redshift defined by the user, and sort in order
    # Turn into a set so that exact matching user-set redshift
    # don't double-up with scrolling ones.
    if (
        (inputs.astro_options.USE_TS_FLUCT or inputs.astro_options.INHOMO_RECO)
        and user_redshifts
        and min(inputs.node_redshifts) > min(user_redshifts)
    ):
        warnings.warn(
            f"minimum node redshift {min(inputs.node_redshifts)} is above output redshift {min(user_redshifts)},"
            + "This may result in strange evolution",
            stacklevel=2,
        )

    zmin_user = min(user_redshifts) if user_redshifts else 0
    needed_nodes = [z for z in inputs.node_redshifts if z > zmin_user]
    redshifts = np.concatenate((needed_nodes, user_redshifts))
    redshifts = np.sort(np.unique(redshifts))[::-1]
    return redshifts.tolist()



"""Tests of the RSDS module."""

import numpy as np
import pytest
from astropy import units

import py21cmfast as p21c
from py21cmfast import InputParameters, run_coeval
from py21cmfast.lightconers import AngularLightconer, RectilinearLightconer
from py21cmfast.rsds import apply_rsds, include_dvdr_in_tau21, rsds_shift
from py21cmfast.wrapper.classy_interface import run_classy


class TestFindRequiredLightconeLimits:
    """Tests of the find_required_lightcone_limits method."""

    def setup_class(self):
        """Set up the RectilinearLightconer for testing."""
        node_redshifts = p21c.get_logspaced_redshifts(
            min_redshift=6.0, max_redshift=35.0, z_step_factor=1.02
        )
        self.inputs = p21c.InputParameters(
            random_seed=12345, node_redshifts=node_redshifts
        )

        self.lcner = RectilinearLightconer.between_redshifts(
            min_redshift=self.inputs.node_redshifts[-1] + 0.5,
            max_redshift=self.inputs.node_redshifts[0] - 0.5,
            resolution=self.inputs.simulation_options.cell_size,
            quantities=("brightness_temp",),
        )

    def test_limits_are_reasonable(self):
        """Test that the limits returned by find_required_lightcone_limits are reasonable."""
        classy = run_classy(output="vTk")
        limits = self.lcner.find_required_lightcone_limits(classy, inputs=self.inputs)
        assert len(limits) == 2
        assert limits[0] < limits[1]
        assert limits[0] <= self.lcner.lc_distances.min()
        assert limits[1] >= self.lcner.lc_distances.max()
        assert limits[0] > self.lcner.lc_distances.min() - 2 * units.Mpc
        assert limits[1] < self.lcner.lc_distances.max() + 2 * units.Mpc


def test_coeval_rsds(ic, default_input_struct_ts, cache):
    """Test rsds on coeval boxes."""
    coeval = run_coeval(
        initial_conditions=ic,
        inputs=default_input_struct_ts,
        cache=cache,
    )
    box_rsd = coeval[0].apply_velocity_corrections()
    assert box_rsd.shape == coeval[0].brightness_temperature.brightness_temp.shape

    box_rsd = coeval[0].include_dvdr_in_tau21()
    assert box_rsd.shape == coeval[0].brightness_temperature.brightness_temp.shape

    box_rsd = coeval[0].apply_rsds()
    assert box_rsd.shape == coeval[0].brightness_temperature.brightness_temp.shape


def test_bad_coeval_inputs(default_input_struct, cache):
    coeval = run_coeval(
        inputs=default_input_struct.evolve_input_structs(KEEP_3D_VELOCITIES=False),
        out_redshifts=8.0,
        cache=cache,
    )
    with pytest.raises(ValueError, match="You asked for axis ="):
        coeval[0].apply_velocity_corrections(
            axis="x"
        )  # fails because KEEP_3D_VELOCITIES = False
    with pytest.raises(ValueError, match="You asked for axis ="):
        coeval[0].include_dvdr_in_tau21(
            axis="x"
        )  # fails because KEEP_3D_VELOCITIES = False
    with pytest.raises(ValueError, match="You asked for axis ="):
        coeval[0].apply_rsds(axis="x")  # fails because KEEP_3D_VELOCITIES = False
    with pytest.raises(ValueError, match=r"`axis` can only be `x`, `y` or `z`."):
        coeval[0].apply_velocity_corrections(axis="t")
    with pytest.raises(ValueError, match=r"`axis` can only be `x`, `y` or `z`."):
        coeval[0].include_dvdr_in_tau21(axis="t")
    with pytest.raises(ValueError, match=r"`axis` can only be `x`, `y` or `z`."):
        coeval[0].apply_rsds(axis="t")


def test_bad_lightconer_inputs(default_input_struct_ts):
    lcner = RectilinearLightconer.between_redshifts(
        min_redshift=default_input_struct_ts.node_redshifts[-1],
        max_redshift=default_input_struct_ts.node_redshifts[0],
        resolution=default_input_struct_ts.simulation_options.cell_size,
        quantities=("brightness_temp",),
    )
    with pytest.raises(
        ValueError,
        match="The lightcone redshifts are not compatible with the given redshift",
    ):
        p21c.run_lightcone(lightconer=lcner, inputs=default_input_struct_ts)

    lcner = RectilinearLightconer.between_redshifts(
        min_redshift=default_input_struct_ts.node_redshifts[-1],
        max_redshift=default_input_struct_ts.simulation_options.Z_HEAT_MAX,
        resolution=default_input_struct_ts.simulation_options.cell_size,
        quantities=("brightness_temp",),
    )
    with pytest.raises(
        ValueError, match="You have set apply_rsds to True with node redshifts between"
    ):
        p21c.run_lightcone(
            lightconer=lcner, inputs=default_input_struct_ts, apply_rsds=True
        )


class TestRSDsShift:
    """Tests of the rsds_shift function."""

    @pytest.mark.parametrize("n_rsd_subcells", [1, 2, 4, 5])
    def test_mass_conservation(self, n_rsd_subcells):
        """Test that sum along LOS is perserved in cloud in cell for a periodic box."""
        nslices = 10
        nangles = 5
        rng = np.random.default_rng(12345)
        shape = (nslices, nangles)
        box_in = rng.random(shape)
        los_displacement = rng.random(shape) * units.pixel
        box_out = rsds_shift(
            field=box_in,
            los_displacement=los_displacement,
            n_rsd_subcells=n_rsd_subcells,
            periodic=True,
        )

        sum_in = np.sum(box_in, axis=0)
        sum1 = np.sum(box_out, axis=0)
        np.testing.assert_allclose(sum_in, sum1)

    @pytest.mark.parametrize("n_rsd_subcells", [1, 2])
    @pytest.mark.parametrize("velocity", [-10, -1, 0, 1, 10])
    def test_integer_shift(self, n_rsd_subcells: int, velocity: int):
        """Test that cloud in cell results in a shifted box, for an integer velocity and a periodic box."""
        nslices = 10
        nangles = 5
        rng = np.random.default_rng(12345)
        box_in = rng.random((nslices, nangles))
        los_displacement = velocity * np.ones_like(box_in) * units.pixel
        box_out = rsds_shift(
            field=box_in,
            los_displacement=los_displacement,
            n_rsd_subcells=n_rsd_subcells,
            periodic=True,
        )

        box_in_shifted = np.roll(box_in, velocity, axis=0)
        np.testing.assert_allclose(box_out, box_in_shifted)

    @pytest.mark.parametrize("n_rsd_subcells", [1, 2, 5])
    def test_non_periodic_large_displacement(self, n_rsd_subcells: int):
        """Test that a very large displacement results in all mass leaving the box."""
        nslices = 10
        nangles = 5
        box_in = np.ones((nslices, nangles))
        los_displacement = nslices * 2 * np.ones_like(box_in) * units.pixel

        box_out = rsds_shift(
            field=box_in,
            los_displacement=los_displacement,
            periodic=False,
            n_rsd_subcells=n_rsd_subcells,
        )
        np.testing.assert_allclose(box_out, 0)


class TestComputeRSDs:
    """Tests of include_dvdr_in_tau21 and apply_rsds functions.

    Since these functions are pretty much just a wrapper around rsds_shift,
    we mostly test for case-behaviour (e.g. testing for errors raised),
    rather than for exactness of the output, which is done in TestRSDsShift.
    """

    def setup_class(self):
        """Set up the arrays re-used throughout the class."""
        self.nslc = 10
        self.nang = 5

        self.bt3d = np.ones((self.nang, self.nang, self.nslc))
        self.bt2d = np.ones((self.nang, self.nslc))
        self.vel3d = np.ones_like(self.bt3d)
        self.vel2d = np.ones_like(self.bt2d)
        self.inputs = InputParameters.from_template(
            "simple", node_redshifts=[6, 7, 8, 40], random_seed=1
        )

    @pytest.mark.parametrize("periodic", [True, False])
    def test_bad_inputs(self, periodic: bool | None):
        """Test that bad inputs raise good errors."""
        with pytest.raises(ValueError, match="tau_21 is not provided"):
            include_dvdr_in_tau21(
                brightness_temp=self.bt3d,
                los_velocity=self.vel3d,
                redshifts=6.0,
                inputs=self.inputs.evolve_input_structs(USE_TS_FLUCT=True),
                tau_21=None,
                periodic=periodic,
            )
        with pytest.raises(
            ValueError,
            match="Redshifts must be a float or array with the same size as number of LoS slices",
        ):
            include_dvdr_in_tau21(
                brightness_temp=self.bt3d,
                los_velocity=self.vel3d,
                redshifts=[6.0, 8.0],
                inputs=self.inputs,
                periodic=periodic,
            )
        with pytest.raises(
            ValueError,
            match="brightness_temp must be an array with the same shape as los_velocity",
        ):
            include_dvdr_in_tau21(
                brightness_temp=self.bt3d[:, :, :-1],
                los_velocity=self.vel3d,
                redshifts=6.0,
                inputs=self.inputs,
                periodic=periodic,
            )
        with pytest.raises(ValueError, match="field must have at least 2 slices"):
            apply_rsds(
                field=self.bt3d[:, :, 0].reshape(self.nang, self.nang, 1),
                los_velocity=self.vel3d[:, :, 0].reshape(self.nang, self.nang, 1),
                redshifts=6.0,
                inputs=self.inputs,
                periodic=periodic,
            )
        with pytest.raises(
            ValueError,
            match="Redshifts must be a float or array with the same size as number of LoS slices",
        ):
            apply_rsds(
                field=self.bt3d,
                los_velocity=self.vel3d,
                redshifts=[6.0, 8.0],
                inputs=self.inputs,
                periodic=periodic,
            )
        with pytest.raises(
            ValueError,
            match="field must be an array with the same shape as los_displacement",
        ):
            apply_rsds(
                field=self.bt3d[:, :, :-1],
                los_velocity=self.vel3d,
                redshifts=6.0,
                inputs=self.inputs,
                periodic=periodic,
            )
        with pytest.raises(ValueError, match="n_rsd_subcells must be an integer"):
            apply_rsds(
                field=self.bt3d,
                los_velocity=self.vel3d,
                redshifts=6.0,
                inputs=self.inputs,
                periodic=periodic,
                n_rsd_subcells=2.5,
            )

    @pytest.mark.parametrize("periodic", [True, False])
    def test_2d_ok(self, periodic: bool | None):
        """Test that 2D brightness temp arrays are OK."""
        rng = np.random.default_rng(1019)
        nslc = 12
        bt3d = rng.uniform(-100, 30, size=(3, 3, nslc))
        vel = rng.uniform(-1.5, 1.5, size=bt3d.shape)

        box_out_3d = apply_rsds(
            field=bt3d,
            los_velocity=vel,
            redshifts=8.0,
            inputs=self.inputs,
            periodic=periodic,
        )

        box_out_2d = apply_rsds(
            field=bt3d.reshape((-1, nslc)),
            los_velocity=vel.reshape((-1, nslc)),
            redshifts=8.0,
            inputs=self.inputs,
            periodic=periodic,
        )

        np.testing.assert_allclose(box_out_3d.flatten(), box_out_2d.flatten())


@pytest.mark.parametrize("lcner", ["rect", "ang"])
def test_new_rsd_lightcones(cache, lcner):
    """Test that new lightcones are added to output."""
    inputs = p21c.InputParameters(
        random_seed=12345,
        node_redshifts=p21c.get_logspaced_redshifts(
            min_redshift=34.0, max_redshift=35.0, z_step_factor=1.02
        ),
    ).evolve_input_structs(
        BOX_LEN=15,
        DIM=45,
        HII_DIM=10,
        N_THREADS=1,
        USE_TS_FLUCT=True,
        KEEP_3D_VELOCITIES=True,
    )
    if lcner == "rect":
        lightconer = RectilinearLightconer.between_redshifts(
            min_redshift=inputs.node_redshifts[-1] + 0.5,
            max_redshift=inputs.node_redshifts[0] - 0.5,
            resolution=inputs.simulation_options.cell_size,
            cosmo=inputs.cosmo_params.cosmo,
            quantities=("brightness_temp", "density"),
        )
    else:
        lightconer = AngularLightconer.like_rectilinear(
            max_redshift=inputs.node_redshifts[0] - 0.5,
            simulation_options=inputs.simulation_options,
            match_at_z=inputs.node_redshifts[-1] + 0.5,
            cosmo=inputs.cosmo_params.cosmo,
            quantities=("brightness_temp", "density"),
        )

    lightcone = p21c.run_lightcone(
        lightconer=lightconer,
        inputs=inputs,
        cache=cache,
        include_dvdr_in_tau21=True,
        apply_rsds=True,
    )
    assert "tau_21" in lightcone.lightcones
    assert "density_with_rsds" in lightcone.lightcones

    assert not np.allclose(
        lightcone.lightcones["brightness_temp"],
        lightcone.lightcones["brightness_temp_with_rsds"],
    )
    assert not np.allclose(
        lightcone.lightcones["density"], lightcone.lightcones["density_with_rsds"]
    )
    assert not np.allclose(lightcone.lightcones["tau_21"], 0)



"""Tests of the templates module."""

from itertools import chain
from pathlib import Path

import pytest

from py21cmfast import _templates as tmpl
from py21cmfast.wrapper.inputs import CosmoTables, InputParameters, InputStruct

_TEMPLATES = tmpl.list_templates()
_ALL_ALIASES = list(chain.from_iterable(t["aliases"] for t in _TEMPLATES))


class TestListTemplates:
    """Tests of the list_templates function."""

    def setup_class(self):
        """Set up the class."""
        self.templates = _TEMPLATES

    def test_output_form(self):
        """Test that list_templates() returns the correct format of list."""
        assert isinstance(self.templates, list)
        assert all(isinstance(x, dict) for x in self.templates)

    def test_existence_of_files(self):
        """Ensure that all template files exist."""
        assert all((tmpl.TEMPLATE_PATH / t["file"]).exists() for t in self.templates)

    def test_uniqueness_of_names(self):
        """Test that all names in the templates are unique."""
        allnames = [t["name"] for t in self.templates]
        assert len(set(allnames)) == len(allnames)

    def test_uniqueness_of_aliases(self):
        """Test that all aliases in the templates are unique."""
        assert len(set(_ALL_ALIASES)) == len(_ALL_ALIASES)


class TestLoadTemplateFile:
    """Tests of the load_template_file function."""

    @pytest.mark.parametrize("template", _ALL_ALIASES)
    def test_load_str(self, template: str):
        """Test that all templates only have correct InputStruct keys in them."""
        out = tmpl.load_template_file(template)
        possible = set(InputStruct._subclasses.keys())
        assert all(key in possible for key in out)

    def test_load_path(self, tmp_path: Path):
        """Test that loading from an explicit TOML path works."""
        pth = tmp_path / "tmp.toml"
        with pth.open("w") as fl:
            fl.write(
                """
                [SimulationOptions]
                BOX_LEN = 100
                """
            )

        out = tmpl.load_template_file(pth)
        assert out["SimulationOptions"]["BOX_LEN"] == 100

    def test_non_existent(self):
        """Test that loading a non-existent template errors."""
        with pytest.raises(ValueError, match="Template VIRIDISSUCKS not found on-disk"):
            tmpl.load_template_file("VIRIDISSUCKS")


class TestCreateParamsFromTemplate:
    """Tests of the create_params_from_template function."""

    def test_string_template_equiv_to_list_of_one_string(self):
        """Test that loading a string is the same as loading a len-1 list."""
        inp1 = tmpl.create_params_from_template("simple")
        inp2 = tmpl.create_params_from_template(["simple"])
        assert inp1 == inp2

    @pytest.mark.parametrize("template", _ALL_ALIASES)
    def test_each_template_works(self, template: str):
        """Test that each template creates a dict of input structs."""
        dct = tmpl.create_params_from_template(template)
        cosmo_tables = dct.pop("cosmo_tables")
        assert all(isinstance(v, InputStruct) for v in dct.values())
        assert isinstance(cosmo_tables, CosmoTables)

    def test_adding_kwargs(self):
        """Test that also specifying loose params does set those params."""
        dct = tmpl.create_params_from_template("simple", HII_DIM=1000)
        assert dct["simulation_options"].HII_DIM == 1000


class TestWriteTemplate:
    """Tests of the write_template function."""

    def test_defaults_minimal(self, tmp_path):
        """Test that writing a minimal-mode default file has nothing in it."""
        inputs = InputParameters.from_template("default", random_seed=1)
        pth = tmp_path / "tmp.toml"
        tmpl.write_template(inputs, pth, mode="minimal")

        with pth.open("r") as fl:
            lines = [line for line in fl.readlines() if line]

        assert len(lines) == 2
        assert "This file was generated by py21cmfast" in lines[0]
        assert "Created on" in lines[1]

    @pytest.mark.parametrize("template", _ALL_ALIASES)
    @pytest.mark.parametrize("mode", ["full", "minimal"])
    def test_roundtrip(self, template, tmp_path: Path, mode: str):
        """Test that writing then reading a template gives the same answer."""
        # Some templates require running CLASS. We set K_MAX_FOR_CLASS to be small so the test won't take too long
        inputs = InputParameters.from_template(
            template, random_seed=1, K_MAX_FOR_CLASS=1.0
        )
        pth = tmp_path / "tmp.toml"
        tmpl.write_template(inputs, pth, mode=mode)

        new = InputParameters.from_template(pth, random_seed=1, K_MAX_FOR_CLASS=1.0)
        assert new == inputs



"""Module defining HDF5 backends for reading/writing output structures.

These functions are those used by default in the caching system of 21cmFAST.
In the future, it is possible that other backends might be implemented.

As of version 4, all cache files from 21cmFAST will have the following heirarchical
structure::

    /attrs/
      |-- 21cmFAST-version
      |-- [redshift]
    /<OutputStructName>/
      /InputParameters/
        /attrs/
          |-- 21cmFAST-version
          |-- random_seed
        /simulation_options/
        /matter_options/
        /cosmo_params/
        /astro_options/
        /astro_params/
        /node_redshifts/
      /OutputFields/
        /attrs/
          |-- [primitive_field_1]
          |-- [primitive_field_2]
          |-- [...]
        /[field_1]/
        /[field_2]/
        /.../

"""

import warnings
from pathlib import Path
from typing import Any

import attrs
import h5py
import numpy as np

from .. import __version__
from ..input_serialization import deserialize_inputs, prepare_inputs_for_serialization
from ..wrapper import outputs as ostruct
from ..wrapper.arrays import H5Backend
from ..wrapper.arraystate import ArrayState
from ..wrapper.inputs import InputParameters


def hdf5_to_dict(grp: h5py.Group) -> dict[str, Any]:
    """Load all data from an HDF5 Group into a dict.

    Essentially the same as toml.load() but for HDF5.
    """
    out = dict(grp.attrs)

    for k, v in grp.items():
        if isinstance(v, h5py.Group):
            out[k] = hdf5_to_dict(v)
        else:
            out[k] = v[()]

    return out


class HDF5FileStructureError(ValueError):
    """An error in the structure of an HDF5 file for 21cmFAST."""


def write_output_to_hdf5(
    output: ostruct.OutputStruct,
    path: Path,
    group: str | None = None,
    mode: str = "w",
):
    """
    Write an output struct in standard HDF5 format.

    Parameters
    ----------
    output
        The OutputStruct to write.
    path : Path
        The path to write the output struct to.
    group : str, optional
        The HDF5 group into which to write the object. By default, this is the root.
    mode : str
        The mode in which to open the file.
    """
    if not all(v.state.is_computed for v in output.arrays.values()):
        raise ValueError(
            "Not all boxes have been computed (or maybe some have been purged). Cannot write."
            f"Non-computed boxes: {[k for k, v in output.arrays.items() if not v.state.is_computed]}. "
            f"Computed boxes: {[k for k, v in output.arrays.items() if v.state.is_computed]}"
        )

    path = Path(path)
    if not path.parent.exists():
        path.parent.mkdir(exist_ok=True, parents=True)

    with h5py.File(path, mode) as fl:
        if group is not None:
            group = fl[group] if group in fl else fl.create_group(group)
        else:
            group = fl

        group.attrs["21cmFAST-version"] = __version__
        group = group.create_group(output._name)

        if hasattr(output, "redshift"):
            group.attrs["redshift"] = output.redshift

        write_outputs_to_group(output, group)
        _write_inputs_to_group(output.inputs, group)


def _write_inputs_to_group(
    inputs: InputParameters, group: h5py.Group | h5py.File | str | Path
) -> None:
    """Write an InputParameters object into a cache file.

    Here we are careful to close the file only if a raw Path is given, and keep it open
    if a h5py.File/Group is given (since then this is likely being called from another
    function that is also writing other objects to the same file).

    Parameters
    ----------
    inputs
        The input parameters object to write.
    group : h5py.Group | h5py.File | str | Path
        The group or file into which to write the inputs. Note that a new group called
        "InputParameters" will be created inside this group/file.
    """
    if not isinstance(group, h5py.Group):
        with h5py.File(group, "a") as fl:
            _write_inputs_to_group(inputs, fl)
        return

    grp = group.create_group("InputParameters")

    # Write 21cmFAST version to the file
    grp.attrs["21cmFAST-version"] = __version__

    inputsdct = prepare_inputs_for_serialization(
        inputs, mode="full", only_structs=True, camel=False
    )

    # Write the input structs. Note that all the "work" for converting attributes
    # to appropriate values is done in the serialization method above, not here.
    for name, dct in inputsdct.items():
        _grp = grp.create_group(name)
        for key, val in dct.items():
            try:
                _grp.attrs[key] = val
            except TypeError as e:
                if isinstance(val, dict):
                    # A "second layer" of recursion is needed since CosmoTables has an attribute that is itself a non-primitive class (Table1D)
                    _grp_dict = _grp.create_group(key)
                    for key_dict, val_dict in val.items():
                        _grp_dict.attrs[key_dict] = val_dict
                else:
                    raise TypeError(
                        f"key {key} with value {val} is not able to be written to HDF5 attrs!"
                    ) from e

    grp.attrs["random_seed"] = inputs.random_seed
    grp["node_redshifts"] = (
        h5py.Empty(None)
        if inputs.node_redshifts is None
        else np.array(inputs.node_redshifts)
    )


def write_outputs_to_group(
    output: ostruct.OutputStruct, group: h5py.Group | h5py.File | str | Path
):
    """
    Write the compute fields of an OutputStruct to a particular HDF5 subgroup.

    Here we are careful to close the file only if a raw Path is given, and keep it open
    if a h5py.File/Group is given (since then this is likely being called from another
    function that is also writing other objects to the same file).

    Parameters
    ----------
    output
        The OutputStruct to write.
    group
        The HDF5 group into which to write the object. A new group "OutputFields" will
        be created inside this group/file.
    """
    need_to_close = False
    if isinstance(group, str | Path):
        file = h5py.File(group, "r")
        group = file
        need_to_close = True

    # Go through all fields in this struct, and save
    group = group.create_group("OutputFields")

    # First make sure we have everything in memory
    output.load_all()

    for k, array in output.arrays.items():
        new = array.written_to_disk(H5Backend(group.file.filename, f"{group.name}/{k}"))
        setattr(output, k, new)

    for k in output.struct.primitive_fields:
        try:
            group.attrs[k] = getattr(output, k)
        except TypeError as e:
            raise TypeError(f"Error writing attribute {k} to HDF5") from e

    group.attrs["21cmFAST-version"] = __version__

    if need_to_close:
        file.close()


def read_output_struct(
    path: Path, group: str = "/", struct: str | None = None, safe: bool = True
) -> ostruct.OutputStruct:
    """
    Read an output struct from an HDF5 file.

    Parameters
    ----------
    path : Path
        The path to the HDF5 file.
    group : str, optional
        A path within the HDF5 heirarchy to the top-level of the OutputStruct. This is
        usually the root of the file.
    struct
        A string specifying the kind of OutputStruct to read (e.g. InitialConditions).
        Generally, this does not need to be provided, as cache files contain just a
        single output struct.
    safe
        Whether to read the file in "safe" mode. If True, keys found in the file that
        are not valid attributes of the struct will raise an exception. If False, only
        a warning will be raised.

    Returns
    -------
    OutputStruct
        An OutputStruct that is contained in the cache file.
    """
    with h5py.File(path, "r") as fl:
        group = fl[group]

        if struct is None:
            if len(group.keys()) > 1:
                raise HDF5FileStructureError(
                    f"Multiple structs found in {path}:{group}"
                )
            else:
                struct = next(iter(group.keys()))
            group = group[struct]

        elif struct in group:
            group = group[struct]
        else:
            raise KeyError(f"struct {struct} not found in the H5DF group {group}")

        kls = getattr(ostruct, struct)
        assert "InputParameters" in group
        assert "OutputFields" in group

        redshift = group.attrs.get("redshift")
        inputs = read_inputs(group["InputParameters"], safe=safe)
        out = _read_outputs(
            group["OutputFields"], struct=kls, redshift=redshift, inputs=inputs
        )

    return out


def read_inputs(
    group: h5py.Group | Path | h5py.File, safe: bool = True
) -> InputParameters:
    """Read the InputParameters from a cache file.

    Parameters
    ----------
    group : h5py.Group | Path | h5py.File
        A file, or HDF5 Group within a file, to read the input parameters from.
    safe : bool, optional
        If in safe mode, errors will be raised if keys exist in the file that are not
        valid attributes of the InputParameters. Otherwise, only warnings will be raised.

    Returns
    -------
    inputs : InputParameters
        The input parameters contained in the file.
    """
    close_after = False
    if isinstance(group, Path):
        file = h5py.File(group, "r")
        if "InputParameters" in file:
            group = file["InputParameters"]
        elif len(file.keys()) > 1:
            raise HDF5FileStructureError(
                f"Multiple sub-groups found in {group}, none of them 'InputParameters'"
            )
        else:
            groupname = next(iter(file.keys()))
            group = file[groupname]["InputParameters"]
        close_after = True
    elif isinstance(group, h5py.File):
        group = group["InputParameters"]

    file_version = group.attrs.get("21cmFAST-version", None)
    if file_version is None:
        raise NotImplementedError(
            f"The file {group.file.filename} is not a valid 21cmFAST v4 file."
        )

    if file_version > __version__:
        warnings.warn(
            f"File created with a newer version {file_version} of 21cmFAST than this {__version__}. "
            f"Reading may break. Consider updating 21cmFAST.",
            stacklevel=2,
        )

    out = _read_inputs_v4(group, safe=safe)

    if close_after:
        file.close()

    return out


def _read_inputs_v4(group: h5py.Group, safe: bool = True):
    # Read the input parameter dictionaries from file.
    kwargs = hdf5_to_dict(group)
    del kwargs["21cmFAST-version"]

    # The node_redshifts and random_seed are treated differently.
    node_redshifts = kwargs.pop("node_redshifts")
    random_seed = kwargs.pop("random_seed")

    kwargs = deserialize_inputs(kwargs, safe=safe)
    return InputParameters(
        node_redshifts=node_redshifts, random_seed=random_seed, **kwargs
    )


def _read_outputs(
    group: h5py.Group,
    struct: type[ostruct.OutputStruct],
    redshift: float | None,
    inputs: InputParameters,
):
    file_version = group.attrs.get("21cmFAST-version", None)

    if file_version is None:
        raise NotImplementedError(
            f"The file {group.file.filename} is not a valid 21cmFAST v4 file."
        )

    if file_version > __version__:
        warnings.warn(
            f"File created with a newer version of 21cmFAST than this. Reading may break. Consider updating 21cmFAST to at least {file_version}",
            stacklevel=2,
        )

    return _read_outputs_v4(group, struct, redshift, inputs)


def _read_outputs_v4(
    group: h5py.Group,
    struct: type[ostruct.OutputStruct],
    redshift: float | None,
    inputs: InputParameters,
):
    # First read other attributes that are not arrays.
    kwargs = dict(group.attrs)
    del kwargs["21cmFAST-version"]

    if redshift is not None:
        kwargs["redshift"] = redshift

    # Create the object with those attributes.
    obj = struct.new(inputs, **kwargs)

    # Now go and make sure all the arrays exist in the file, and have the correct shape.
    # We don't actually read these right now, we just make pointers to the file.
    for name, array in obj.arrays.items():
        if name not in group:
            raise HDF5FileStructureError(
                f"Required Array {name} not found in {group}. This file is not valid."
            )

        dataset = group[name]
        if dataset.shape != array.shape:
            raise HDF5FileStructureError(
                f"Array {name} has shape {dataset.shape} in the file {group.file.filename}, but requires shape {array.shape}"
            )

        # We don't check dtype because it can be usually safely cast.

        setattr(
            obj,
            name,
            attrs.evolve(
                array,
                state=ArrayState(on_disk=True),
                cache_backend=H5Backend(
                    path=group.file.filename, dataset=group.name + f"/{name}"
                ),
            ),
        )

    return obj



"""Module for accounting redshift space distortions."""

from collections.abc import Sequence

import numpy as np
from astropy import cosmology, units
from classy import Class
from cosmotile.cic import cloud_in_cell_los
from scipy import fft
from scipy.interpolate import RegularGridInterpolator

from .wrapper.classy_interface import compute_rms
from .wrapper.inputs import InputParameters


def include_dvdr_in_tau21(
    brightness_temp: np.ndarray,
    los_velocity: np.ndarray,
    redshifts: np.ndarray | float,
    inputs: InputParameters,
    periodic: bool,
    tau_21: np.ndarray | None = None,
) -> np.ndarray:
    """Include velocity gradient corrections to the brightness temperature field.

    Parameters
    ----------
    brightness_temp
        A box of the brightness temperature, without velocity gradient corrections.
    los_velocity
        Line-of-sight velocities for each cell of brightness_temp, in Mpc/second.
    redshifts
        An array of the redshifts along the los. Can also be a float (could be useful for coeval boxes)
    inputs
        The input parameters corresponding to the box.
    tau_21
        A box of the 21cm optical depth. Not required if inputs.astro_options.USE_TS_FLUCT = False.
    periodic
        Whether to assume periodic boundary conditions along the line-of-sight.

    Returns
    -------
    tb_with_dvdr
        A box of the brightness temperature, with velocity gradient corrections.
    """
    if tau_21 is None and inputs.astro_options.USE_TS_FLUCT:
        raise ValueError(
            "tau_21 is not provided, but inputs.astro_options.USE_TS_FLUCT is True!"
        )
    if hasattr(redshifts, "__len__") and len(redshifts) != brightness_temp.shape[-1]:
        raise ValueError(
            "Redshifts must be a float or array with the same size as number of LoS slices"
        )
    if los_velocity.shape != brightness_temp.shape:
        raise ValueError(
            "brightness_temp must be an array with the same shape as los_velocity"
        )

    N = inputs.simulation_options.HII_DIM

    dx_los = inputs.simulation_options.BOX_LEN / N  # Mpc
    # TODO: currently the gradient is applied w.r.t to the z-axis, even if the user specified a different los-axis. Needs to make it more flexible in the future
    if periodic:
        k_real = fft.rfftfreq(N, dx_los) * 2.0 * np.pi
        k_complex = fft.fftfreq(N, dx_los) * 2.0 * np.pi
        k_vector = np.stack(np.meshgrid(k_complex, k_complex, k_real, indexing="ij"))
        vel_gradient = (
            fft.irfftn(1j * k_vector[-1] * fft.rfftn(los_velocity), s=(N, N, N))
            / units.s
        )
    else:
        vel_gradient = np.gradient(
            los_velocity * units.Mpc / units.s,
            dx_los * units.Mpc,
            axis=-1,
            edge_order=2,
        )

    H = inputs.cosmo_params.cosmo.H(redshifts)

    if not inputs.astro_options.USE_TS_FLUCT:
        # If we don't have spin temperature, we also assume tau_21 << 1 and make the Taylor approximation.
        # Clipping is required so the brightness temperature won't diverge when the gradient term goes to small values
        max_v_deriv = inputs.astro_params.MAX_DVDR * H
        dvdx = np.clip(vel_gradient, -max_v_deriv, max_v_deriv)
        gradient_component = np.abs(1.0 + dvdx / H)
        tb_with_dvdr = brightness_temp / gradient_component.value
    else:
        # We have the spin temperature, and we do *not* make the Taylor approximation
        tb_no_rsds = brightness_temp.copy()
        tau_21 = np.float64(tau_21)
        gradient_component = np.float64(np.abs(1.0 + vel_gradient / H))
        with np.errstate(
            divide="ignore", invalid="ignore"
        ):  # Don't show division by 0 warnings
            gradient_factor = (1.0 - np.exp(-tau_21 / gradient_component)) / (
                1.0 - (np.exp(-tau_21))
            )
        # It doesn't really matter what the rsd_factor is when tau_21=0 because the brightness temperature is zero by definition
        gradient_factor = np.float32(np.where(tau_21 < 1e-10, 1.0, gradient_factor))
        tb_with_dvdr = tb_no_rsds * gradient_factor.value

    return tb_with_dvdr


def apply_rsds(
    field: np.ndarray,
    los_velocity: np.ndarray,
    redshifts: np.ndarray | float,
    inputs: InputParameters,
    periodic: bool,
    n_rsd_subcells: int = 4,
) -> np.ndarray:
    """Apply redshift-space distortions to a field.

    Notes
    -----
    To ensure that we cover all the slices in the field after the velocities have
    been applied, we extrapolate the densities and velocities on either end by the
    maximum velocity offset in the field.
    Then, to ensure we don't pick up cells with zero particles (after displacement),
    we interpolate the slices onto a finer regular grid (in comoving distance) and
    then displace the field on that grid, before interpolating back onto the original
    slices.

    Parameters
    ----------
    field
        A box of the field on which redshift-space distortions shall be applied.
    los_velocity
        Line-of-sight velocities for each cell of brightness_temp, in Mpc/second.
    redshifts
        An array of the redshifts along the los. Can also be a float (could be useful for coeval boxes)
    inputs
        The input parameters corresponding to the box.
    periodic: bool
        Whether to assume periodic boundary conditions along the line-of-sight.
    n_rsd_subcells: int, optional
        The number of subcells into which each cell is divided when redshift space distortions are applied. Default is 4.

    Returns
    -------
    field_with_rsds
        A box of the field, with redshift space distortions.
    """
    if hasattr(redshifts, "__len__") and len(redshifts) != field.shape[-1]:
        raise ValueError(
            "Redshifts must be a float or array with the same size as number of LoS slices"
        )

    H = inputs.cosmo_params.cosmo.H(redshifts)
    los_displacement = los_velocity * units.Mpc / units.s / H
    equiv = units.pixel_scale(inputs.simulation_options.cell_size / units.pixel)
    los_displacement = los_displacement.to(units.pixel, equivalencies=equiv)

    # We transform rectilinear lightcone to be an angular-like lightcone
    field_dimensions = len(field.shape)
    if field_dimensions == 3:
        field = field.reshape((field.shape[0] * field.shape[1], -1))
        los_displacement = los_displacement.reshape(
            (los_displacement.shape[0] * los_displacement.shape[1], -1)
        )

    # Here we move the cells along the line of sight, regardless the geometry (rectilinear or angular)
    field_with_rsds = rsds_shift(
        field=field.T,
        los_displacement=los_displacement.T,
        n_rsd_subcells=n_rsd_subcells,
        periodic=periodic,
    ).T

    # And now we transform back to a rectilinear-like lightcone
    if field_dimensions == 3:
        field_with_rsds = field_with_rsds.reshape(
            (
                int(np.sqrt(field_with_rsds.shape[0])),
                int(np.sqrt(field_with_rsds.shape[0])),
                -1,
            )
        )
    return field_with_rsds


def rsds_shift(
    field: np.ndarray,
    los_displacement: np.ndarray,
    n_rsd_subcells: int = 4,
    periodic: bool = False,
) -> np.ndarray:
    """Shift the cells of a field according to the los displacement.

    Parameters
    ----------
    field
        A box of the field on which redshift-space distortions shall be applied, shape (nslices, ncoords).
    los_displacement
        The line-of-sight "apparent" displacement of the field, in pixel coordinates.
        Equal to ``v / H(z) / cell_size``.
        Positive values are towards the observer, shape ``(nslices, ncoords)``.
    periodic: bool, optioanl
        Whether to assume periodic boundary conditions along the line-of-sight.
    n_rsd_subcells: int, optional
        The number of subcells into which each cell is divided when redshift space distortions are applied. Default is 4.

    Returns
    -------
    field_with_rsds
        A box of the field, with redshift space distortions, shape (nslices, ncoords).
    """
    if field.shape[0] < 2:
        raise ValueError("field must have at least 2 slices")
    if los_displacement.shape != field.shape:
        raise ValueError(
            "field must be an array with the same shape as los_displacement"
        )
    if not isinstance(n_rsd_subcells, int):
        raise ValueError("n_rsd_subcells must be an integer")

    ang_coords = np.arange(field.shape[1])

    distance = np.arange(field.shape[0])

    # We need to extend the grid once as we would like to interpret the displacement values
    # to be associated with the *centers* of the cells, as was previously done in the C code
    distance_plus = np.arange(field.shape[0] + 1)

    if periodic:
        # We extend the grid twice more to account for periodic boundary conditions
        distance_plus_periodic = np.arange(-1, field.shape[0] + 2)
        distance_grid = (distance_plus_periodic[1:] + distance_plus_periodic[:-1]) / 2
        # We also extend the displacement array to have periodic boundary conditions
        first_slice = los_displacement[-1, :].reshape(1, len(ang_coords))
        last_slice = los_displacement[0, :].reshape(1, len(ang_coords))
        los_displacement = np.concatenate(
            (first_slice, los_displacement, last_slice), axis=0
        )

    else:
        distance_grid = (distance_plus[1:] + distance_plus[:-1]) / 2

    fine_field = np.repeat(field, n_rsd_subcells, axis=0) / n_rsd_subcells

    # This is where we shall evaluate the subcells
    distance_fine = np.linspace(
        distance_plus.min(),
        distance_plus.max(),
        1 + n_rsd_subcells * (len(distance_plus) - 1),
    )
    fine_grid = (distance_fine[1:] + distance_fine[:-1]) / 2
    x, y = np.meshgrid(fine_grid, ang_coords, indexing="ij")
    grid = (x.flatten(), y.flatten())
    fine_rsd = RegularGridInterpolator(
        (distance_grid, ang_coords),
        los_displacement * n_rsd_subcells,
        bounds_error=False,
        fill_value=None,
        method="linear",
    )(grid).reshape(x.shape)

    fine_field = cloud_in_cell_los(fine_field, fine_rsd, periodic=periodic)
    # Average the subcells back to the original grid
    return np.sum(
        fine_field.T.reshape(len(ang_coords), len(distance), n_rsd_subcells), axis=-1
    ).T


def estimate_rsd_displacements(
    classy_output: Class,
    cosmo: cosmology,
    redshifts: Sequence[float],
    factor: float = 1.0,
) -> Sequence[units.Quantity]:
    """Estimate the rms of the redshift space distortions displacement field at given redshifts.

    Parameters
    ----------
    classy_output : classy.Class
        An object containing all the information from the CLASS calculation.
    cosmo: astropy.cosmology
        The assumed cosmology
    redshifts: list
        List of the redshifts at which the rms of the displacement field is computed.
    factor: float, optional
        Factor to multiply the rms velocity from CLASS. Default is 1.

    Returns
    -------
    displacements : np.array
        The rms of the RSD displacement field.
    """
    v_rms = (
        compute_rms(classy_output=classy_output, kind="v_b", redshifts=redshifts)
        * factor
    )
    # The multiplication by (1+z)=1/a is because CLASS returns the *proper* velocity field,
    # while we work in *comoving* coordinates
    return v_rms / cosmo.H(redshifts) * (1.0 + redshifts)



"""Test the wrapper functions which access the C-backend, but not though an OutputStruct compute() method."""

import matplotlib as mpl
import numpy as np
import pytest
from hmf import MassFunction

import py21cmfast as p21c
from py21cmfast.wrapper import cfuncs as cf


def test_run_lf():
    inputs = p21c.InputParameters(random_seed=9)
    *_, lf = p21c.compute_luminosity_function(
        inputs=inputs, redshifts=[7, 8, 9], nbins=100
    )
    assert np.all(lf[~np.isnan(lf)] > -30)
    assert lf.shape == (3, 100)

    # Check that memory is in-tact and a second run also works:
    _muv, _mhalo, lf2 = p21c.compute_luminosity_function(
        inputs=inputs, redshifts=[7, 8, 9], nbins=100
    )
    assert lf2.shape == (3, 100)
    assert np.allclose(lf2[~np.isnan(lf2)], lf[~np.isnan(lf)])

    inputs = inputs.from_template("mini", random_seed=9)

    _muv_minih, _mhalo_minih, lf_minih = p21c.compute_luminosity_function(
        redshifts=[7, 8, 9],
        nbins=100,
        component="mcg",
        inputs=inputs,
        mturnovers=[7.0, 7.0, 7.0],
        mturnovers_mini=[5.0, 5.0, 5.0],
    )
    assert np.all(lf_minih[~np.isnan(lf_minih)] > -30)
    assert lf_minih.shape == (3, 100)


def test_run_tau():
    inputs = p21c.InputParameters(random_seed=9)
    tau = p21c.compute_tau(
        redshifts=[7, 8, 9],
        global_xHI=[0.1, 0.2, 0.3],
        inputs=inputs,
        z_re_HeII=3.0,
    )

    assert tau


def test_bad_integral_inputs(default_input_struct):
    # make arrays with different shapes
    redshifts = np.linspace(6, 35, num=20)
    lnM_base = np.linspace(7, 13, num=30)
    densities = np.linspace(-1, 3, num=25)

    with pytest.raises(ValueError, match="the shapes of"):
        cf.integrate_chmf_interval(
            inputs=default_input_struct,
            redshift=redshifts[0],
            lnm_lower=lnM_base,
            lnm_upper=lnM_base[1:],
            cond_values=densities,
        )

    with pytest.raises(ValueError, match="the shapes of"):
        cf.evaluate_inverse_table(
            inputs=default_input_struct,
            redshift=redshifts[0],
            cond_array=densities,
            probabilities=np.ones(len(densities) - 1),
        )

    with pytest.raises(ValueError, match="the shapes of"):
        cf.evaluate_SFRD_z(
            inputs=default_input_struct,
            redshifts=redshifts,
            log10mturns=lnM_base,
        )

    with pytest.raises(ValueError, match="the shapes of"):
        cf.evaluate_Nion_z(
            inputs=default_input_struct,
            redshifts=redshifts,
            log10mturns=lnM_base,
        )

    with pytest.raises(ValueError, match="the shapes of"):
        cf.evaluate_SFRD_cond(
            inputs=default_input_struct,
            redshift=redshifts[0],
            densities=densities,
            log10mturns=lnM_base,
            radius=10.0,
        )

    with pytest.raises(ValueError, match="the shapes of"):
        cf.evaluate_Nion_cond(
            inputs=default_input_struct,
            redshift=redshifts[0],
            densities=densities,
            l10mturns_acg=lnM_base,
            l10mturns_mcg=lnM_base,
            radius=10.0,
        )

    with pytest.raises(ValueError, match="the shapes of"):
        cf.evaluate_Xray_cond(
            inputs=default_input_struct,
            redshift=redshifts[0],
            densities=densities,
            log10mturns=lnM_base,
            radius=10.0,
        )

    with pytest.raises(
        ValueError, match="Halo masses and rng shapes must be identical"
    ):
        cf.convert_halo_properties(
            inputs=default_input_struct,
            redshift=redshifts[0],
            halo_masses=np.zeros(10),
            star_rng=np.zeros(10),
            sfr_rng=np.zeros(10),
            xray_rng=np.zeros(11),
        )


@pytest.mark.parametrize("hmf_model", ["PS", "ST"])
@pytest.mark.parametrize("ps_model", ["EH", "BBKS"])
@pytest.mark.xfail(reason="pending proper comparison between 21cmFAST and hmf")
def test_matterfield_statistics(default_input_struct, hmf_model, ps_model, plt):
    redshift = 8.0
    hmf_map = {
        "PS": "PS",
        "ST": "SMT",
    }
    transfer_map = {
        "EH": "EH_NoBAO",
        "BBKS": "BBKS",
    }

    if ps_model == "BBKS":
        t_params = {"use_sugiyama_baryons": True, "use_liddle_baryons": False}
    else:
        t_params = {}

    if hmf_model == "PS":
        hmf_params = {}
    elif hmf_model == "ST":
        hmf_params = {"a": 0.73, "p": 0.175, "A": 0.353}

    comparison_mf = MassFunction(
        z=redshift,
        Mmin=7,
        Mmax=16,
        hmf_model=hmf_map[hmf_model],
        hmf_params=hmf_params,
        transfer_model=transfer_map[ps_model],
        transfer_params=t_params,
        delta_c=1.68,  # hmf default is 1.686
    )

    inputs = default_input_struct.clone(
        cosmo_params=p21c.CosmoParams.from_astropy(
            comparison_mf.cosmo,
            SIGMA_8=comparison_mf.sigma_8,
            POWER_INDEX=comparison_mf.n,
        ),
    ).evolve_input_structs(
        POWER_SPECTRUM=ps_model,
        HMF=hmf_model,
        USE_INTERPOLATION_TABLES="no-interpolation",
    )
    h = inputs.cosmo_params.cosmo.h

    hmf_vals = cf.return_uhmf_value(
        inputs=inputs,
        redshift=redshift,
        mass_values=comparison_mf.m / h,
    )

    sigma_vals, dsigmasq_vals = cf.evaluate_sigma(
        inputs=inputs,
        masses=comparison_mf.m / h,
    )

    power_vals = cf.get_matter_power_values(
        inputs=inputs,
        k_values=comparison_mf.k * h,
    )

    mass_dens = (
        inputs.cosmo_params.cosmo.critical_density(0).to("M_sun Mpc^-3").value
        * inputs.cosmo_params.cosmo.Om0
    )

    if plt == mpl.pyplot:
        make_matterfield_comparison_plot(
            comparison_mf.m / h,
            comparison_mf.k * h,
            [
                comparison_mf.dndlnm * (h**3),
                comparison_mf._sigma_0,
                -comparison_mf._dlnsdlnm * comparison_mf._sigma_0 / comparison_mf.m * h,
                comparison_mf._power0 / (h**3),
            ],
            [
                hmf_vals * mass_dens,
                sigma_vals,
                -dsigmasq_vals / (2 * sigma_vals),
                power_vals,
            ],
            plt,
        )

    # check matter power spectrum
    np.testing.assert_allclose(power_vals, comparison_mf._power0 / (h**3), rtol=1e-3)
    # check sigma(M)
    np.testing.assert_allclose(sigma_vals, comparison_mf._sigma_0, rtol=1e-3)
    # check dSigma/dM
    np.testing.assert_allclose(
        dsigmasq_vals / (2 * sigma_vals),
        comparison_mf._dlnsdlnm * comparison_mf._sigma_0 / comparison_mf.m * h,
        rtol=1e-3,
    )
    # check mass function
    np.testing.assert_allclose(
        mass_dens * hmf_vals, comparison_mf.dndlnm * (h**3), rtol=1e-3
    )


@pytest.mark.parametrize("hmf_model", ["PS", "ST"])
@pytest.mark.parametrize(
    "ps_model", ["EH", "BBKS", "EFSTATHIOU", "PEEBLES", "WHITE", "CLASS"]
)
def test_hmf_runs(default_input_struct, hmf_model, ps_model):
    mass_range = np.logspace(7, 12, num=64)
    redshift = 8.0

    inputs = default_input_struct.evolve_input_structs(
        POWER_SPECTRUM=ps_model,
        HMF=hmf_model,
    )

    hmf_vals = cf.return_uhmf_value(
        inputs=inputs,
        redshift=redshift,
        mass_values=mass_range,
    )

    assert hmf_vals.shape == (len(mass_range),)
    assert np.all(~np.isnan(hmf_vals))


@pytest.mark.parametrize("hmf_model", ["PS", "ST"])
@pytest.mark.parametrize(
    "ps_model", ["EH", "BBKS", "EFSTATHIOU", "PEEBLES", "WHITE", "CLASS"]
)
def test_chmf_runs(default_input_struct, hmf_model, ps_model):
    delta_range = np.linspace(-1.0, 1.7, num=32)
    condmass_range = np.logspace(8, 13, num=16)
    mass_range = np.logspace(7, 12, num=64)
    redshift = 8.0

    inputs = default_input_struct.evolve_input_structs(
        POWER_SPECTRUM=ps_model,
        HMF=hmf_model,
    )

    hmf_vals = cf.return_chmf_value(
        inputs=inputs,
        redshift=redshift,
        mass_values=mass_range,
        delta_values=delta_range,
        condmass_values=condmass_range,
    )

    assert hmf_vals.shape == (len(delta_range), len(condmass_range), len(mass_range))
    assert np.all(~np.isnan(hmf_vals))


def test_ps_runs(default_input_struct):
    k_values = np.logspace(-3, 1, num=64)

    ps = cf.get_matter_power_values(
        inputs=default_input_struct,
        k_values=k_values,
    )

    assert ps.shape == (len(k_values),)
    assert np.all(ps >= 0.0)

    with pytest.raises(
        ValueError,
        match=r"inputs.matter_options.USE_RELATIVE_VELOCITIES must be True in order to compute the v_cb power spectrum\.",
    ):
        cf.get_vcb_power_values(
            inputs=default_input_struct,
            k_values=k_values,
        )

    ps = cf.get_vcb_power_values(
        inputs=default_input_struct.evolve_input_structs(
            POWER_SPECTRUM="CLASS",
            USE_RELATIVE_VELOCITIES=True,
            K_MAX_FOR_CLASS=1.0,
        ),
        k_values=k_values,
    )

    assert ps.shape == (len(k_values),)
    assert np.all(ps >= 0.0)


def test_ps_with_A_s_and_sigma8(default_input_struct):
    """Test that we get the same power spectrum, regadless if we use A_s or sigma_8."""
    k_values = np.logspace(-3, 1, num=64)

    input_CLASS = default_input_struct.evolve_input_structs(
        POWER_SPECTRUM="CLASS", K_MAX_FOR_CLASS=1.0
    )
    A_s = input_CLASS.cosmo_params.A_s

    ps_sigma8 = cf.get_matter_power_values(
        inputs=input_CLASS,
        k_values=k_values,
    )
    ps_A_s = cf.get_matter_power_values(
        inputs=input_CLASS.evolve_input_structs(A_s=A_s, SIGMA_8=None),
        k_values=k_values,
    )
    np.testing.assert_allclose(
        ps_sigma8,
        ps_A_s,
        rtol=5e-4,
    )


def make_matterfield_comparison_plot(
    x,
    k,
    true,
    test,
    plt,
    **kwargs,
):
    _fig, axs = plt.subplots(nrows=2, ncols=4, figsize=(16, 8))

    for i, row in enumerate(axs):
        for j, ax in enumerate(row):
            ax.set_xscale("log")
            ax.grid()
            if i == 1:
                ax.set_xlabel("Mass") if j < 2 else ax.set_xlabel("k")
                ax.set_ylim(-1, 1)
                ax.axhline(0, color="k", linestyle="--")
            else:
                ax.set_yscale("log")

            if j == 0:
                ax.set_ylabel("dndlnM")
            elif j == 1:
                ax.set_ylabel("Sigma")
            elif j == 2:
                ax.set_ylabel("dSigmadM")
            else:
                ax.set_ylabel("Power Spectrum")

    axs[0, 0].loglog(x, true[0], linestyle="-", label="Truth", **kwargs)
    axs[0, 0].loglog(x, test[0], linestyle=":", linewidth=3, label="Test", **kwargs)
    axs[1, 0].semilogx(x, (test[0] - true[0]) / true[0], **kwargs)
    axs[0, 0].legend()
    axs[0, 0].set_ylim(1e-12, 1e2)

    axs[0, 1].loglog(x, true[1], linestyle="-", label="Truth", **kwargs)
    axs[0, 1].loglog(x, test[1], linestyle=":", linewidth=3, label="Test", **kwargs)
    axs[1, 1].semilogx(x, (test[1] - true[1]) / true[1], **kwargs)

    axs[0, 2].loglog(x, true[2], linestyle="-", label="Truth", **kwargs)
    axs[0, 2].loglog(x, test[2], linestyle=":", linewidth=3, label="Test", **kwargs)
    axs[1, 2].semilogx(x, (test[2] - true[2]) / true[2], **kwargs)

    axs[0, 3].loglog(k, true[3], linestyle="-", label="Truth", **kwargs)
    axs[0, 3].loglog(k, test[3], linestyle=":", linewidth=3, label="Test", **kwargs)
    axs[1, 3].semilogx(k, (test[3] - true[3]) / true[3], **kwargs)



"""Various tests of the compute_initial_conditions() function and InitialConditions class."""

from multiprocessing import cpu_count

import numpy as np
import pytest

import py21cmfast as p21c


@pytest.fixture(scope="module")
def ic_hires(default_input_struct) -> p21c.InitialConditions:
    return p21c.compute_initial_conditions(
        inputs=default_input_struct.evolve_input_structs(PERTURB_ON_HIGH_RES=True),
    )


@pytest.fixture(scope="module")
def ic_from_ic(default_input_struct, ic) -> p21c.InitialConditions:
    """Make initial conditions, given the hires density of an initial condition instance."""
    return p21c.compute_initial_conditions(
        inputs=default_input_struct, initial_density=ic.hires_density.value
    )


@pytest.fixture(scope="module")
def single_pxl_array_mean_zero(default_input_struct) -> np.ndarray:
    """Make a single pixel array, with mean zero."""
    dim = default_input_struct.simulation_options.DIM
    array = -np.ones((dim,) * 3)
    array[dim // 2, dim // 2, dim // 2] = array[0, 0, 0] - array.sum()
    return array


@pytest.fixture(scope="module")
def ic_from_array(
    default_input_struct, single_pxl_array_mean_zero
) -> p21c.InitialConditions:
    """Make initial conditions, given the hires density array of a single pixel."""
    return p21c.compute_initial_conditions(
        inputs=default_input_struct, initial_density=single_pxl_array_mean_zero
    )


def test_box_shape(ic_hires: p21c.InitialConditions, ic: p21c.InitialConditions):
    """Test basic properties of the InitialConditions struct."""
    shape = (35, 35, 35)
    hires_shape = tuple(2 * s for s in shape)
    # test common fields
    for box in (ic, ic_hires):
        assert box.lowres_density.shape == shape
        assert box.hires_density.shape == hires_shape
        assert box.hires_vx_2LPT.shape == hires_shape
        assert box.hires_vy_2LPT.shape == hires_shape
        assert box.hires_vz_2LPT.shape == hires_shape
        assert box.lowres_vcb is None
        assert box.cosmo_params == p21c.CosmoParams()

    # test hires only fields

    assert ic_hires.hires_vx.shape == hires_shape
    assert ic_hires.hires_vy.shape == hires_shape
    assert ic_hires.hires_vz.shape == hires_shape

    assert ic_hires.lowres_vx is None
    assert ic_hires.lowres_vy is None
    assert ic_hires.lowres_vz is None

    assert ic_hires.lowres_vx_2LPT is None
    assert ic_hires.lowres_vy_2LPT is None
    assert ic_hires.lowres_vz_2LPT is None

    # test lowres only fields

    assert ic.lowres_vx.shape == shape
    assert ic.lowres_vy.shape == shape
    assert ic.lowres_vz.shape == shape

    assert ic.lowres_vx_2LPT.shape == shape
    assert ic.lowres_vy_2LPT.shape == shape
    assert ic.lowres_vz_2LPT.shape == shape

    assert ic.hires_vx is None
    assert ic.hires_vy is None
    assert ic.hires_vz is None


def test_modified_cosmo(
    ic: p21c.InitialConditions, default_input_struct: p21c.InputParameters, cache
):
    """Test using a modified cosmology."""
    inputs = default_input_struct.evolve_input_structs(SIGMA_8=0.9)
    ic2 = p21c.compute_initial_conditions(inputs=inputs, cache=cache)

    assert ic2.cosmo_params != ic.cosmo_params
    assert ic2.cosmo_params == inputs.cosmo_params
    assert ic2.cosmo_params.SIGMA_8 == inputs.cosmo_params.SIGMA_8


def test_transfer_function(
    ic: p21c.InitialConditions, default_input_struct: p21c.InputParameters, cache
):
    """Test using a modified transfer function."""
    inputs = default_input_struct.evolve_input_structs(POWER_SPECTRUM="CLASS")
    ic2 = p21c.compute_initial_conditions(inputs=inputs, cache=cache)
    hrd2 = ic2.hires_density.value
    hrd = ic.hires_density.value

    rmsnew = np.sqrt(np.mean(hrd2**2))
    rmsdelta = np.sqrt(np.mean((hrd2 - hrd) ** 2))
    assert rmsdelta < rmsnew
    assert rmsnew > 0.0
    assert not np.allclose(hrd2, hrd)


def test_relvels():
    """Test for relative velocity initial conditions."""
    inputs = p21c.InputParameters(random_seed=1).evolve_input_structs(
        HII_DIM=100,
        DIM=300,
        BOX_LEN=300,
        POWER_SPECTRUM="CLASS",
        USE_RELATIVE_VELOCITIES=True,
        N_THREADS=cpu_count(),  # To make this one a bit faster.
    )
    ic = p21c.compute_initial_conditions(inputs=inputs)

    vcbrms_lowres = np.sqrt(np.mean(ic.lowres_vcb.value**2))
    vcbavg_lowres = np.mean(ic.lowres_vcb.value)

    # we test the lowres box
    # rms should be about 30 km/s for LCDM, so we check it is finite and not far off
    # the average should be 0.92*vrms, since it follows a maxwell boltzmann
    assert vcbrms_lowres > 20.0
    assert vcbrms_lowres < 40.0
    assert vcbavg_lowres < 0.97 * vcbrms_lowres
    assert vcbavg_lowres > 0.88 * vcbrms_lowres


@pytest.mark.parametrize(
    "name",
    [
        "lowres_density",
        "hires_density",
        "lowres_vx",
        "lowres_vy",
        "lowres_vz",
        "lowres_vx_2LPT",
        "lowres_vy_2LPT",
        "lowres_vz_2LPT",
    ],
)
def test_initial_density_array(
    ic: p21c.InitialConditions,
    ic_from_ic: p21c.InitialConditions,
    ic_from_array: p21c.InitialConditions,
    single_pxl_array_mean_zero: np.ndarray,
    name: str,
):
    """Test the functionality with the initial_density argument."""
    # Test that the hires_density arrays are exactly the same (by definition)
    assert np.all(ic_from_ic.hires_density.value == ic.hires_density.value)

    # Test that the other arrays are close (numerical differences exist due to FFT-IFFT)
    np.testing.assert_allclose(
        getattr(ic, name).value, getattr(ic_from_ic, name).value, atol=1e-5, rtol=0.0
    )

    # Test the array we use actually has mean zero
    assert single_pxl_array_mean_zero.mean() == 0.0

    # Test that the hires_density array is exactly our single pixel array input
    assert np.all(ic_from_array.hires_density.value == single_pxl_array_mean_zero)

    # Test that the arrays are different between the original ic and the ic we got from array
    assert not np.allclose(
        getattr(ic, name).value, getattr(ic_from_array, name).value, atol=1e-5, rtol=0.0
    )


def test_bad_initial_density_array(
    default_input_struct: p21c.InitialConditions, single_pxl_array_mean_zero: np.ndarray
):
    """Test bad/weird initial_density array."""
    # Run initial conditions with hires density box that has non-zero mean, just to throw the relevant warning
    ic_non_zero = p21c.compute_initial_conditions(
        inputs=default_input_struct,
        initial_density=np.ones_like(single_pxl_array_mean_zero),
    )
    assert isinstance(ic_non_zero, p21c.InitialConditions)

    with pytest.raises(
        ValueError,
        match="The shape of your high resolution initial_density is not consistent with inputs!",
    ):
        p21c.compute_initial_conditions(
            inputs=default_input_struct,
            initial_density=np.zeros((single_pxl_array_mean_zero.shape[0] * 2,) * 3),
        )



"""Low-level wrapping utilities and functions for the core C-code.

The main wrapper for the underlying 21cmFAST C-code.

The module provides both low- and high-level wrappers, using the very low-level machinery
in :mod:`~py21cmfast._utils`, and the convenient input and output structures from
:mod:`~py21cmfast.inputs` and :mod:`~py21cmfast.outputs`.

This module provides a number of:

* Low-level functions which simplify calling the background C functions which populate
  these output objects given the input classes.
* High-level functions which provide the most efficient and simplest way to generate the
  most commonly desired outputs.

**Low-level functions**

The low-level functions provided here ease the production of the aforementioned output
objects. Functions exist for each low-level C routine, which have been decoupled as far
as possible. So, functions exist to create :func:`initial_conditions`,
:func:`perturb_field`, :class:`ionize_box` and so on. Creating a brightness temperature
box (often the desired final output) would generally require calling each of these in
turn, as each depends on the result of a previous function. Nevertheless, each function
has the capability of generating the required previous outputs on-the-fly, so one can
instantly call :func:`ionize_box` and get a self-consistent result. Doing so, while
convenient, is sometimes not *efficient*, especially when using inhomogeneous
recombinations or the spin temperature field, which intrinsically require consistent
evolution of the ionization field through redshift. In these cases, for best efficiency
it is recommended to either use a customised manual approach to calling these low-level
functions, or to call a higher-level function which optimizes this process.

Finally, note that :mod:`py21cmfast` attempts to optimize the production of the large amount of
data via on-disk caching. By default, if a previous set of data has been computed using
the current input parameters, it will be read-in from a caching repository and returned
directly. This behaviour can be tuned in any of the low-level (or high-level) functions
by setting the `write`, `direc`, `regenerate` and `match_seed` parameters (see docs for
:func:`initial_conditions` for details). The function :func:`~query_cache` can be used
to search the cache, and return empty datasets corresponding to each (and these can then be
filled with the data merely by calling ``.read()`` on any data set). Conversely, a
specific data set can be read and returned as a proper output object by calling the
:func:`~py21cmfast.cache_tools.readbox` function.


**High-level functions**

As previously mentioned, calling the low-level functions in some cases is non-optimal,
especially when full evolution of the field is required, and thus iteration through a
series of redshift. In addition, while :class:`InitialConditions` and
:class:`PerturbedField` are necessary intermediate data, it is *usually* the resulting
brightness temperature which is of most interest, and it is easier to not have to worry
about the intermediate steps explicitly. For these typical use-cases, two high-level
functions are available: :func:`run_coeval` and :func:`run_lightcone`, whose purpose
should be self-explanatory. These will optimally run all necessary intermediate
steps (using cached results by default if possible) and return all datasets of interest.


Examples
--------
A typical example of using this module would be the following.

>>> import py21cmfast as p21

Get coeval cubes at redshift 7,8 and 9, without spin temperature or inhomogeneous
recombinations:

>>> coeval = p21.run_coeval(
>>>     redshift=[7,8,9],
>>>     cosmo_params=p21.CosmoParams(hlittle=0.7),
>>>     simulation_options=p21.SimulationOptions(HII_DIM=100)
>>> )

Get coeval cubes at the same redshift, with both spin temperature and inhomogeneous
recombinations, pulled from the natural evolution of the fields:

>>> all_boxes = p21.run_coeval(
>>>                 redshift=[7,8,9],
>>>                 simulation_options=p21.SimulationOptions(HII_DIM=100),
>>>                 astro_options=p21.AstroOptions(INHOMO_RECO=True),
>>>                 do_spin_temp=True
>>>             )

Get a self-consistent lightcone defined between z1 and z2 (`z_step_factor` changes the
logarithmic steps between redshift that are actually evaluated, which are then
interpolated onto the lightcone cells):

>>> lightcone = p21.run_lightcone(redshift=z2, max_redshift=z2, z_step_factor=1.03)
"""



"""Produce performance plots for run_global_evolution."""

import time
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import tyro

import py21cmfast as p21c

output_dir = (
    Path(__file__).parent.parent / "images" / "performance" / "global_evolution"
)

ALL_TEMPLATES = ["const-zeta", "latest", "fixed-halos", "minihalos"]

X_ARRAYS = {
    "N_THREADS": np.arange(1, 11),
    "ZPRIME_STEP_FACTOR": [1.02, 1.04, 1.08, 1.1, 1.15, 1.2],
    "Z_HEAT_MAX": np.arange(10.0, 40.0, 5.0),
}


def run_global_evolution(inputs):
    """Run global evolution given inputs, and return runtime in seconds."""
    start = time.time()
    p21c.run_global_evolution(inputs=inputs)
    end = time.time()
    return end - start


def measure_global_evolution_runtime(template, **kwargs):
    """Measure runtime of run_global_evolution given a template and kwargs."""
    zmax = kwargs.get("Z_HEAT_MAX", 35.0)
    zstep_factor = kwargs.get("ZPRIME_STEP_FACTOR", 1.02)

    inputs = p21c.InputParameters.from_template(
        template, random_seed=1234, **kwargs
    ).with_logspaced_redshifts(
        zmax=zmax,
        zstep_factor=zstep_factor,
    )

    runtime = run_global_evolution(inputs)
    return runtime


def plot_and_save_global_evolution_runtime(
    x_array,
    kwarg,
    templates=ALL_TEMPLATES,
):
    """Plot run_global_evolution runtime as a function of a key word argument, as given by the x_array, and save the plot."""
    fig, ax = plt.subplots(1, 1, figsize=(8, 8))
    for template in templates:
        runtime_list = []
        for x in x_array:
            kwargs = {kwarg: x, "USE_TS_FLUCT": True}
            runtime = measure_global_evolution_runtime(template, **kwargs)
            runtime_list.append(runtime)
        ax.plot(x_array, runtime_list, label=template, marker="o")
    ax.set_xlabel(kwarg)
    ax.set_ylabel("Runtime (sec)")
    ax.legend()
    fig.savefig(output_dir / f"global_evolution_runtime_vs_{kwarg}.png")


def go(kwargs=X_ARRAYS.keys()):
    """Plot and save global evolution runtime for each kwarg in kwargs."""
    for kwarg in kwargs:
        plot_and_save_global_evolution_runtime(
            x_array=X_ARRAYS[kwarg],
            kwarg=kwarg,
        )


if __name__ == "__main__":
    tyro.cli(go)



"""
Tests of single-field low-level functions.

These are designed to be unit-tests of the wrapper functionality. They do not test for
correctness of simulations, but whether different parameter options work/don't work as
intended.
"""

import numpy as np
import pytest

import py21cmfast as p21c
from py21cmfast import (
    BrightnessTemp,
    HaloBox,
    HaloCatalog,
    InitialConditions,
    IonizedBox,
    OutputCache,
    PerturbedField,
    TsBox,
)


@pytest.fixture(scope="module")
def ic_newseed(default_input_struct, cache: p21c.OutputCache):
    return p21c.compute_initial_conditions(
        inputs=default_input_struct.clone(random_seed=33), write=True, cache=cache
    )


@pytest.fixture(scope="module")
def perturb_field_lowz(ic: InitialConditions, low_redshift: float, cache: OutputCache):
    """A default perturb_field."""
    return p21c.perturb_field(
        redshift=low_redshift,
        initial_conditions=ic,
        write=True,
        cache=cache,
    )


@pytest.fixture(scope="module")
def ionize_box_lowz(
    ic: InitialConditions,
    perturb_field_lowz: p21c.PerturbedField,
    cache: OutputCache,
):
    """A default ionize_box at lower redshift."""
    return p21c.compute_ionization_field(
        initial_conditions=ic,
        perturbed_field=perturb_field_lowz,
        write=True,
        cache=cache,
    )


@pytest.fixture(scope="module")
def spin_temp_evolution(ic: InitialConditions, default_input_struct_ts: TsBox, cache):
    """An example spin temperature evolution."""
    scrollz = default_input_struct_ts.node_redshifts
    st_prev = None
    outputs = []
    for z in scrollz:
        pt = p21c.perturb_field(
            redshift=z,
            initial_conditions=ic,
            inputs=default_input_struct_ts,
            cache=cache,
        )
        st = p21c.compute_spin_temperature(
            initial_conditions=ic,
            perturbed_field=pt,
            previous_spin_temp=st_prev,
            inputs=default_input_struct_ts,
            cache=cache,
        )
        outputs.append(
            {
                "redshift": z,
                "perturbed_field": pt,
                "spin_temp": st,
            }
        )
        st_prev = st

    return outputs


def test_pf_unnamed_param():
    """Try using an un-named parameter."""
    with pytest.raises(TypeError):
        p21c.perturb_field(7)


def test_perturb_field_ic(perturbed_field, ic):
    # this will run perturb_field again,
    # it should produce exactly the same as the default perturb_field since it has
    # the same seed.
    pf = p21c.perturb_field(
        redshift=perturbed_field.redshift,
        initial_conditions=ic,
        regenerate=True,
    )

    assert pf.density.shape == ic.lowres_density.shape
    assert pf.cosmo_params == ic.cosmo_params
    assert pf.simulation_options == ic.simulation_options
    assert not np.all(pf.density == 0)

    assert pf.simulation_options == perturbed_field.simulation_options
    assert pf.cosmo_params == perturbed_field.cosmo_params

    assert pf == perturbed_field


def test_new_seeds(
    ic_newseed,
    perturb_field_lowz,
    ionize_box_lowz,
    default_input_struct,
    cache,
):
    # Perturbed Field
    pf = p21c.perturb_field(
        redshift=perturb_field_lowz.redshift, initial_conditions=ic_newseed, cache=cache
    )

    # we didn't write it, and this has a different seed
    assert cache.find_existing(pf) is None
    assert pf.random_seed != perturb_field_lowz.random_seed
    assert not np.all(pf.density.value == perturb_field_lowz.density.value)

    # Ionization Box
    with pytest.raises(
        ValueError,
        match="InputParameters in perturbed_field do not match those in initial_conditions",
    ):
        p21c.compute_ionization_field(
            initial_conditions=ic_newseed,
            perturbed_field=perturb_field_lowz,
            cache=cache,
        )

    ib = p21c.compute_ionization_field(
        initial_conditions=ic_newseed, perturbed_field=pf, cache=cache
    )

    # we didn't write it, and this has a different seed
    assert cache.find_existing(ib) is None
    assert ib.random_seed != ionize_box_lowz.random_seed
    assert not np.all(
        ib.neutral_fraction.value == ionize_box_lowz.neutral_fraction.value
    )


def test_ib_from_pf(perturbed_field, ic, cache):
    ib = p21c.compute_ionization_field(
        initial_conditions=ic, perturbed_field=perturbed_field, cache=cache
    )
    assert ib.redshift == perturbed_field.redshift
    assert ib.inputs == perturbed_field.inputs


def test_ib_bad_st(ic, default_input_struct, perturbed_field, redshift, cache):
    with pytest.raises(TypeError, match="spin_temp should be of type TsBox"):
        p21c.compute_ionization_field(
            inputs=default_input_struct,
            initial_conditions=ic,
            perturbed_field=perturbed_field,
            spin_temp=ic,
            cache=cache,
        )


def test_bt(
    ionize_box, default_input_struct, spin_temp_evolution, perturbed_field, cache
):
    curr_st = spin_temp_evolution[-1]["spin_temp"]

    # this will fail because ionized_box was not created with spin temperature.
    with pytest.raises(
        ValueError,
        match="InputParameters in spin_temp do not match those in ionized_box",
    ):
        p21c.brightness_temperature(
            ionized_box=ionize_box,
            perturbed_field=perturbed_field,
            spin_temp=curr_st,
            cache=cache,
        )

    bt = p21c.brightness_temperature(
        ionized_box=ionize_box, perturbed_field=perturbed_field, cache=cache
    )

    assert bt.inputs == perturbed_field.inputs


def test_coeval_against_direct(
    redshift: float,
    ic: p21c.InitialConditions,
    perturbed_field: p21c.PerturbedField,
    ionize_box: p21c.IonizedBox,
    cache,
):
    [coeval] = p21c.run_coeval(
        out_redshifts=redshift, initial_conditions=ic, cache=cache
    )

    assert coeval.initial_conditions == ic
    assert coeval.perturbed_field == perturbed_field
    assert coeval.ionized_box == ionize_box


def test_parameter_override(
    ic: p21c.InitialConditions,
    default_input_struct: p21c.InputParameters,
    ionize_box: p21c.IonizedBox,
    perturbed_field: p21c.PerturbedField,
):
    """Ensure that we use the correct parameters for all calculations.

    Tests compatible but unequal parameter sets
    """
    inputs_changenodes = default_input_struct.clone(node_redshifts=(12, 10, 8))

    pf = p21c.perturb_field(
        redshift=12.0, initial_conditions=ic, inputs=inputs_changenodes
    )

    assert isinstance(pf, p21c.PerturbedField)
    assert pf.inputs == inputs_changenodes
    assert pf.inputs != ic.inputs
    assert pf != perturbed_field

    inputs_changeastro = inputs_changenodes.evolve_input_structs(F_STAR10=-3.0)

    ib = p21c.compute_ionization_field(
        initial_conditions=ic,
        perturbed_field=pf,
        inputs=inputs_changeastro,
        write=False,
    )

    assert isinstance(ib, p21c.IonizedBox)
    assert ib.inputs == inputs_changeastro
    assert ib.inputs != pf.inputs
    assert ib != ionize_box


def test_incompatible_parameters(
    ic: p21c.InitialConditions,
    default_input_struct: p21c.InputParameters,
    ionize_box: p21c.IonizedBox,
    perturbed_field: p21c.PerturbedField,
):
    """Ensure that we throw errors when incompatible parameters are given."""
    df_dim = default_input_struct.simulation_options.DIM
    inputs_changedim = default_input_struct.evolve_input_structs(DIM=df_dim + 1)
    with pytest.raises(
        ValueError, match="InputParameters in InitialConditions do not match those in"
    ):
        p21c.perturb_field(
            redshift=10.0, initial_conditions=ic, inputs=inputs_changedim
        )

    inputs_changenodes = default_input_struct.clone(node_redshifts=(12, 10, 8))
    with pytest.raises(
        ValueError, match="InputParameters in PerturbedField do not match those in"
    ):
        p21c.compute_ionization_field(
            initial_conditions=ic,
            perturbed_field=perturbed_field,
            inputs=inputs_changenodes,
            write=False,
        )


def test_using_cached_halo_catalog(ic, test_direc):
    """Test whether the C-based memory in halo fields is cached correctly.

    Prior to v3.1 this was segfaulting, so this test ensure that this behaviour does
    not regress.
    """
    cache = OutputCache(test_direc)
    halo_catalog = p21c.determine_halo_catalog(
        redshift=10.0, initial_conditions=ic, write=True, cache=cache
    )

    pt_halos = p21c.perturb_halo_catalog(
        initial_conditions=ic,
        halo_catalog=halo_catalog,
        write=True,
        cache=cache,
    )

    # Now get the halo field again at the same redshift -- should be cached
    new_halo_catalog = p21c.determine_halo_catalog(
        redshift=10.0,
        initial_conditions=ic,
        write=False,
        regenerate=False,
    )

    new_pt_halos = p21c.perturb_halo_catalog(
        initial_conditions=ic,
        halo_catalog=new_halo_catalog,
        write=False,
        regenerate=False,
    )

    np.testing.assert_allclose(
        new_halo_catalog.halo_masses.value, halo_catalog.halo_masses.value
    )
    np.testing.assert_allclose(
        pt_halos.halo_coords.value, new_pt_halos.halo_coords.value
    )


def test_incompatible_redshifts(default_input_struct, ic):
    """Test whether bad redshifts are handled correctly."""
    inputs = default_input_struct.clone(node_redshifts=np.array([16.0, 14.0, 12.0]))

    # generate some PetrurbedFields to use
    # NOTE: the checks still happen even if the field is not used due to flags
    ptb_list = [
        p21c.perturb_field(
            initial_conditions=ic, redshift=z, inputs=inputs, write=False
        )
        for z in inputs.node_redshifts
    ]

    kw = {
        "initial_conditions": ic,
        "inputs": inputs,
    }
    # try passing the current redshift == previous
    with pytest.raises(ValueError, match="Incompatible redshifts with inputs"):
        p21c.compute_ionization_field(
            perturbed_field=ptb_list[1],
            previous_perturbed_field=ptb_list[1],
            **kw,
        )

    # try passing the previous redshift < current
    with pytest.raises(ValueError, match="Incompatible redshifts with inputs"):
        p21c.compute_ionization_field(
            perturbed_field=ptb_list[1],
            previous_ionize_box=ptb_list[2],
            **kw,
        )

    # try skipping a step
    with pytest.raises(ValueError, match="Incompatible redshifts with inputs"):
        p21c.compute_ionization_field(
            perturbed_field=ptb_list[2],
            previous_perturbed_field=ptb_list[0],
            **kw,
        )


def test_photoncons_backend_error(redshift, default_input_struct, ic):
    """Test whether the error is raised when you try a photoncons run without proper setup."""
    inputs = default_input_struct.evolve_input_structs(PHOTON_CONS_TYPE="z-photoncons")

    # first test if the error occurs with no inputs
    with pytest.raises(
        ValueError, match="Photon conservation is needed but not initialised"
    ):
        p21c.perturb_field(redshift=redshift, initial_conditions=ic, inputs=inputs)

    # test if the error occurs with the wrong inputs
    p21c.setup_photon_cons(initial_conditions=ic, inputs=inputs)
    with pytest.raises(
        ValueError, match="Photon conservation is needed but not initialised"
    ):
        p21c.perturb_field(
            redshift=redshift,
            initial_conditions=ic,
            inputs=inputs.evolve_input_structs(PHOTON_CONS_TYPE="f-photoncons"),
        )

    # finally test if the calibration works when set up correctly
    p21c.perturb_field(
        redshift=redshift,
        initial_conditions=ic,
        inputs=inputs,
    )


def test_global_properties(
    default_input_struct, ionize_box, spin_temp_evolution, perturbed_field
):
    """Test the global properties which exist in some OutputStruct subclasses."""
    st_new = TsBox.new(inputs=default_input_struct, redshift=10.0)
    with pytest.raises(AttributeError, match="global_Ts is not defined"):
        _ = st_new.global_Ts

    with pytest.raises(AttributeError, match="global_Tk is not defined"):
        _ = st_new.global_Tk

    with pytest.raises(AttributeError, match="global_x_e is not defined"):
        _ = st_new.global_x_e

    ion_new = IonizedBox.new(inputs=default_input_struct, redshift=10.0)
    with pytest.raises(AttributeError, match="global_xH is not defined"):
        _ = ion_new.global_xH

    bt_new = BrightnessTemp.new(inputs=default_input_struct, redshift=10.0)
    with pytest.raises(AttributeError, match="global_Tb is not defined"):
        _ = bt_new.global_Tb

    curr_st = spin_temp_evolution[-1]["spin_temp"]
    assert curr_st.global_Ts == np.mean(curr_st.get("spin_temperature"))
    assert curr_st.global_Tk == np.mean(curr_st.get("kinetic_temp_neutral"))
    assert curr_st.global_x_e == np.mean(curr_st.get("xray_ionised_fraction"))

    assert ionize_box.global_xH == np.mean(ionize_box.get("neutral_fraction"))

    bt = p21c.brightness_temperature(
        ionized_box=ionize_box, perturbed_field=perturbed_field
    )
    assert bt.global_Tb == np.mean(bt.get("brightness_temp"))


def test_bad_input_structs(default_input_struct_ts):
    """Test that we raise errors when required input structs are omitted."""
    # setting parameters for the maximum number of fields required
    test_inputs = default_input_struct_ts.evolve_input_structs(
        USE_MINI_HALOS=True,
        SOURCE_MODEL="CHMF-SAMPLER",
        INHOMO_RECO=True,
    ).clone(node_redshifts=(35.0, 11.0, 10.0))

    # We don't need to compute since we arent testing a successful run
    ic = InitialConditions.new(inputs=test_inputs)
    hf = HaloCatalog.new(redshift=10.0, inputs=test_inputs, buffer_size=1)
    hb = HaloBox.new(redshift=10.0, inputs=test_inputs)
    pt = PerturbedField.new(redshift=10.0, inputs=test_inputs)
    pt_p = PerturbedField.new(redshift=11.0, inputs=test_inputs)
    st = TsBox.new(redshift=10.0, inputs=test_inputs)
    st_p = TsBox.new(redshift=11.0, inputs=test_inputs)
    ib_p = IonizedBox.new(redshift=11.0, inputs=test_inputs)

    # PerturbedHaloCatalog
    with pytest.raises(
        ValueError, match="Below Z_HEAT_MAX you must specify the previous_spin_temp"
    ):
        p21c.perturb_halo_catalog(
            initial_conditions=ic,
            halo_catalog=hf,
            previous_ionize_box=ib_p,
        )
    with pytest.raises(
        ValueError, match="Below Z_HEAT_MAX you must specify the previous_ionize_box"
    ):
        p21c.perturb_halo_catalog(
            initial_conditions=ic,
            halo_catalog=hf,
            previous_spin_temp=st_p,
        )

    # HaloBox
    with pytest.raises(
        ValueError, match="You must provide halo_catalog for SOURCE_MODEL"
    ):
        p21c.compute_halo_grid(
            redshift=10.0,
            initial_conditions=ic,
            previous_ionize_box=ib_p,
            previous_spin_temp=st_p,
        )
    with pytest.raises(
        ValueError, match="Below Z_HEAT_MAX you must specify the previous_spin_temp"
    ):
        p21c.compute_halo_grid(
            redshift=10.0,
            initial_conditions=ic,
            halo_catalog=hf,
            previous_ionize_box=ib_p,
        )
    with pytest.raises(
        ValueError, match="Below Z_HEAT_MAX you must specify the previous_spin_temp"
    ):
        p21c.compute_halo_grid(
            redshift=10.0,
            initial_conditions=ic,
            halo_catalog=hf,
        )

    # TsBox
    with pytest.raises(
        ValueError,
        match="xray_source_box is required for SOURCE_MODEL",
    ):
        p21c.compute_spin_temperature(
            initial_conditions=ic,
            perturbed_field=pt,
            previous_spin_temp=st_p,
        )

    # IonizedBox
    with pytest.raises(
        ValueError,
        match="You need to provide a previous ionized box when redshift < Z_HEAT_MAX",
    ):
        p21c.compute_ionization_field(
            initial_conditions=ic,
            perturbed_field=pt,
            previous_perturbed_field=pt_p,
            halobox=hb,
            spin_temp=st,
        )
    with pytest.raises(
        ValueError,
        match="You need to provide a previous perturbed field when redshift < Z_HEAT_MAX",
    ):
        p21c.compute_ionization_field(
            initial_conditions=ic,
            perturbed_field=pt,
            halobox=hb,
            previous_ionized_box=ib_p,
            spin_temp=st,
        )
    with pytest.raises(ValueError, match="A HaloBox must be provided for SOURCE_MODEL"):
        p21c.compute_ionization_field(
            initial_conditions=ic,
            perturbed_field=pt,
            previous_perturbed_field=pt_p,
            previous_ionized_box=ib_p,
            spin_temp=st,
        )
    with pytest.raises(
        ValueError, match="No spin temperature box given but USE_TS_FLUCT=True"
    ):
        p21c.compute_ionization_field(
            initial_conditions=ic,
            perturbed_field=pt,
            previous_perturbed_field=pt_p,
            halobox=hb,
            previous_ionized_box=ib_p,
        )
